/*
 ******************************************************************************
 *
 * This file is a part of the IBIS ICM Parser.
 * Copyright (C) 2003-2005 Kelly C. Green <kgreen22@mindspring.com>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 ******************************************************************************
 *
 * $URL: file:///home/kgreen/ibis-svn-root/icmchk/trunk/src/icm-1.0-yacc.y $
 *
 * $Id: icm-1.0-yacc.y 121 2005-03-12 18:07:00Z kgreen $
 *
 ******************************************************************************
 */




%debug
%locations
%verbose
%error-verbose

%{

/*
 ******************************************************************************
 *
 * This file is a part of the IBIS ICM Parser.
 * Copyright (C) 2003-2005 Kelly C. Green <kgreen22@mindspring.com>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 ******************************************************************************
 *
 * $URL: file:///home/kgreen/ibis-svn-root/icmchk/trunk/src/icm-1.0-yacc.y $
 *
 * $Id: icm-1.0-yacc.y 121 2005-03-12 18:07:00Z kgreen $
 *
 ******************************************************************************
 */

#if ! (defined(__GNUC__) || defined(__attribute__))
#define __attribute__(a) /* a */
#endif
  
  static char *SVNID __attribute__((unused)) = "$Id: icm-1.0-yacc.y 121 2005-03-12 18:07:00Z kgreen $";
  
#include "icm-impl.h"

  /*
   * Redefine the parser's entry point to distinguish it from future revisions
   */

#define yyparse icm_1_0_parse

  /*
   * parser local functions
   */
 
  static char *choice PROTO((YYLTYPE *loc, char *in, char *choices));
 
  static void validate_icm_header  PROTO((icm_header_t *header));
  static void validate_icm_model   PROTO((icm_model_t *model));
  static void validate_icm_family  PROTO((icm_family_t *family));
  static void validate_icm_section PROTO((icm_section_t *section));
  static void validate_icm         PROTO((icm_t *icm));
 
  static void walk_family_tree     PROTO((icm_family_t *family, icm_model_t *model, 
					  icm_pin_map_t **pinmap, icm_tree_path_t **tree));
 
  static int  section_is_diagonal  PROTO((icm_section_t *section));
  static int  matrix_is_diagonal   PROTO((icm_matrix_t *matrix));
 
#ifdef YYDEBUG  
  static int yydebug = 1;  /* turn on parser debugging */
#endif
 
  /*
   * action macros to manipulate the lexical input stream
   *
   * FLUSH_KEYWORD    <-- advance input to next [keyword].
   * FLUSH_LINE       <-- advance input to next '\n'.
   *
   * The look-ahead token is also cleared, but the error state is NOT.
   * If recovery is complete, follow up with yyerrok.
   */
  
#define FLUSH_KEYWORD { yylex_flush_keyword(); yyclearin; }
#define FLUSH_LINE    { yylex_flush_line(); yyclearin; }
  
%}

/*
 * include core token definitions (via m4)
 */

/*
 ******************************************************************************
 *
 * This file is a part of the IBIS ICM Parser.
 * Copyright (C) 2003-2005 Kelly C. Green <kgreen22@mindspring.com>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 ******************************************************************************
 *
 * $URL: file:///home/kgreen/ibis-svn-root/icmchk/trunk/src/icm-1.0-yacc.y $
 *
 * $Id: icm-1.0-yacc.y 121 2005-03-12 18:07:00Z kgreen $
 *
 ******************************************************************************
 */

/*
 * Core datatype and token definitions that are shared with the 
 * lexical analyzer and common to all versions of the parser
 *
 * These definitions are held in a single file to ensure they
 * are included identically in all necessary source files.
 */

%union {
  char letter;
  char *str;
  long integer;
  double real;
  icm_t *icm;
  icm_header_t *header;
  icm_family_t *family;
  icm_model_t *model;
  icm_model_t **model_list;
  icm_tree_path_t *tree_path;
  icm_node_path_t **node_path;
  icm_node_path_t *node_path_entry;
  icm_swath_entry_t *swath_entry;
  icm_swath_entry_t **swath_list;
  icm_pin_map_t *pin_map;
  icm_pin_list_entry_t **pin_list;
  icm_pin_list_entry_t *pin_list_entry;
  icm_node_map_t *node_map;
  icm_node_list_entry_t *node_list_entry;
  char **nodes;
  icm_section_t *section;
  icm_port_assignment_entry_t *port_assignment_entry;
  icm_port_assignment_entry_t **port_assignment;
  icm_matrix_t *matrix;
  icm_matrix_value_t **matrix_value;
}

/*
 * Termianl targets shared with the lexical anaylzer
 */

%token <str> FIRST_KEYWORD

%token <str> BEGIN_HEADER
%token <str> ICM_VER
%token <str> FILE_NAME
%token <str> FILE_REV
%token <str> DATE
%token <str> SOURCE
%token <str> NOTES
%token <str> DISCLAIMER
%token <str> COPYRIGHT
%token <str> SUPPORT
%token <str> REDISTRIBUTION
%token <str> REDISTRIBUTION_TEXT
%token <str> END_HEADER
%token <str> BEGIN_ICM_FAMILY
%token <str> MANUFACTURER
%token <str> ICM_FAMILY_DESCRIPTION
%token <str> ICM_MODEL_LIST
%token <str> BEGIN_ICM_MODEL
%token <str> ICM_MODEL_DESCRIPTION
%token <str> TREE_PATH_DESCRIPTION
%token <str> NODAL_PATH_DESCRIPTION
%token <str> ICM_SWATH_PARAMETERS
%token <str> ICM_SWATH_PIN_NUMBERS
%token <str> END_ICM_MODEL
%token <str> ICM_PIN_MAP
%token <str> ICM_NODE_MAP
%token <str> END_ICM_FAMILY
%token <str> BEGIN_ICM_SECTION
%token <str> DERIVATION_METHOD
%token <str> RESISTANCE_MATRIX
%token <str> INDUCTANCE_MATRIX
%token <str> CONDUCTANCE_MATRIX
%token <str> CAPACITANCE_MATRIX
%token <str> BANDWIDTH
%token <str> FREQUENCY
%token <str> ROW
%token <str> ICM_S_PARAMETER
%token <str> END_ICM_SECTION
%token <str> END
%token <str> UNKNOWN_KEYWORD   /* anything surrounded by [] that isn't one of the above */

%token <str> LAST_KEYWORD


%token <str> ICM_model_type
%token <str> SGR
%token <real> Ref_impedance_EQ
%token <str> Model_pinmap
%token <str> Side
%token <str> Fork
%token <str> Endfork
%token <str> Section
%token <str> Mult_EQ
%token <str> Len_EQ
%token <str> Model_nodemap
%token <str> N_section
%token <str> Left_edge_EQ
%token <str> Right_edge_EQ
%token <str> Top_edge_EQ
%token <str> Bottom_edge_EQ
%token <str> Pin_order
%token <str> Num_of_rows_EQ
%token <str> Num_of_columns_EQ
%token <str> Pin_list
%token <str> Sparse_matrix
%token <str> Diagonal_matrix
%token <str> Banded_matrix
%token <str> Full_matrix
%token <str> File_name
%token <str> Port_assignment

%token <str>    NODE_NAME         /* alphanumeric symbol */
%token <str>    WORD              /* a space-delineated word */
%token <str>    NUMBER            /* a number */
%token <str>    TEXT_STRING       /* a text string */
%token <str>    TEXT_BLOCK        /* a text block */
%token <str>    FILENAME          /* a file name string */
%token <str>    SGR_RATIO         /* a signal to ground ratio 'INT:INT' */
%token <letter> END_OF_LINE       /* a line termination sequence */

/*
 * Single character tokens, defined only to build up the parser's name table.
 * Without these declarations, parse errors involving these tokens
 * refer to them as an '$undefined' token, which does not
 * result in a particularly illuminating error message.
 */

%token <letter> ' '
%token <letter> '!'
%token <letter> '"'
%token <letter> '#'
%token <letter> '$'
%token <letter> '%'
%token <letter> '&'
%token <letter> '.'
%token <letter> ';'
%token <letter> '<'
%token <letter> '='
%token <letter> '>'
%token <letter> '?'
%token <letter> '@'
%token <letter> '['
%token <letter> '\\'
%token <letter> ']'
%token <letter> '^'
%token <letter> '_'
%token <letter> '`'
%token <letter> '{'
%token <letter> '|'
%token <letter> '}'
%token <letter> '~'


/*
 * Nonterminal targets that collect intermediate parse state.
 * These are the equivalent of local variables.
 */

%type <str> newline
%type <str> node_name
%type <str> word
%type <str> sgr_ratio
%type <str> text_string
%type <str> text_block
%type <real> real_number
%type <integer> integer
%type <str> filename
%type <icm> icm
%type <icm> icm_body
%type <str> end
%type <header> header_part
%type <header> header_body
%type <str> end_header
%type <str> redist_type
%type <family> family_part
%type <family> family_body
%type <str> end_family
%type <model_list> model_list
%type <model> model_entry
%type <str> mating
%type <real> slew
%type <str> image
%type <str> model_type
%type <model> model
%type <model> model_body
%type <str> end_model
%type <tree_path> tree_body
%type <tree_path> model_pinmap
%type <tree_path> model_section
%type <tree_path> interconnect
%type <tree_path> fork
%type <tree_path> fork_body
%type <node_path> nodal_body
%type <node_path_entry> model_nodemap
%type <node_path_entry> n_section
%type <nodes> node_list
%type <nodes> list_of_names
%type <swath_entry> swath_pin_entry
%type <swath_list> swath_pin_list
%type <pin_map> pin_map
%type <str> pin_order
%type <pin_list> pin_list
%type <pin_list_entry> pin_list_entry
%type <node_map> node_map
%type <section> section_part
%type <section> section_body
%type <str> section_end
%type <str> derivation_method
%type <matrix> rlgc_matrix
%type <matrix> matrix
%type <matrix_value> normal_value
%type <matrix_value> normal_entry
%type <matrix_value> diagonal_value
%type <matrix_value> sparse_value
%type <matrix_value> sparse_entry
%type <port_assignment> port_assignment_list
%type <port_assignment_entry> port_assignment_entry

%start icm  /* This is the parser's goal */

%%

/*
 * Newline rule:  
 *
 * exists pretty much only for readability.
 *
 */

newline:
END_OF_LINE {
};


/*
 * These targets exist to get around the fact that the icm language
 * is insufficiently regular for the lexical analyzer to make type
 * determinations statelessly.
 *
 * The lexical analyzer must therefore either retain a significant amount 
 * of state (very bad idea that), or the parser, which already has the 
 * necessary state, must provide contextual hints to the lexical anaylzer 
 * for it to correctly interpret the input stream.
 *
 * The latter is by far the cleanest solution.  (Yes, I did try both ways.)
 * Building state into the lexical analyzer is simply untenable.
 *
 * KLUDGE ALERT:  If the parser must use a look-ahead to reduce one
 * of these targets, the hint WILL NOT WORK, as the look-ahead occurs
 * first.  You can get around this by directly setting the hint with
 * an appropriately placed hint.  Fortunatly, the need for
 * this is quite rare.
 *
 * The hint non-terminals defined below estimate their location
 * to the location returned by the last call to YYLEX.  If they
 * don't do this, their location is row/col 0/0, which can really
 * foul up the location calculation for non-terminals that 
 * reference them.  This is not wholly accurate, but it is better
 * than nothing if you forget to provide a better estimate as
 * the hinted targets below do.
 * 
 * In general, whereever a hint is used, you should look at the
 * effect it will have on location.  The targets below may
 * have leading white space removed that the hint does not
 * otherwise know about.
 */

node_name_hint: /* empty */ { 
  yylex_hint(H_NODE_NAME); 
  @$.first_line   = @$.last_line   = yylloc.last_line;
  @$.first_column = @$.last_column = yylloc.last_column + 1;
};

node_name:
node_name_hint NODE_NAME {
  $$ = $2;
  @$ = @2;
}|
node_name_hint error {
  $$ = _icm_string("");
  _icm_msg(DEBUG, &@1, "Assuming NODE_NAME is ''.");
  _icm_msg(INFO, &@1, "Node names consist of alpha, numeric, and '_' characters only.");
};


word_hint: /* empty */ { 
  yylex_hint(H_WORD); @$ = yylloc; 
  @$.first_line   = @$.last_line   = yylloc.last_line;
  @$.first_column = @$.last_column = yylloc.last_column + 1;
};

word:
word_hint WORD { /* this is actually the parsers default interpretaton */
  $$ = $2;
  @$ = @2;
}|
word_hint error {
  $$ = _icm_string("");
  _icm_msg(DEBUG, &@1, "Assuming WORD is ''.");
  _icm_msg(INFO, &@1, "Words are white space delineated sequences of any dark printing characters.");
};


sgr_ratio_hint: /* empty */ { 
  yylex_hint(H_SGR); @$ = yylloc; 
  @$.first_line   = @$.last_line   = yylloc.last_line;
  @$.first_column = @$.last_column = yylloc.last_column + 1;
};

sgr_ratio:
sgr_ratio_hint SGR_RATIO {
  char *sep = strchr($2, ':');          /* guaranteed to be in form INT:INT */
  long sigs;
  long gnds;
  long greatest;
  char *lptr;

  sigs = strtol($2, &lptr, 10);
  gnds = strtol(&sep[1], &lptr, 10);

  if (sigs <= 0 || gnds <= 0) _icm_msg(ERROR, &@2, "An SGR ratio must be composed of two positive, non-zero integers.");

  /* try to find the greatest common divisor, complain if we can */

  sigs = labs(sigs);
  gnds = labs(gnds);
  greatest = gnds;
  if (gnds > sigs) greatest = sigs;

  while (greatest) {
    ldiv_t result_1 = ldiv(gnds, greatest);
    ldiv_t result_2 = ldiv(sigs, greatest);
    if ((0 == result_1.rem) && (0 == result_2.rem)) break;
    greatest -= 1;
  }

  if (1 < greatest) {
    _icm_msg(WARNING, &@2, "SGR '%s' has a common divisor of '%d'.  Use '%u:%u'.", $2, greatest, sigs/greatest, gnds/greatest);
  }

  $$ = $2;
  @$ = @2;
}|
sgr_ratio_hint error {
  $$ = _icm_string("1:1");
  _icm_msg(DEBUG, &@1, "Assuming SGR is '1:1'.");
  _icm_msg(INFO, &@1, "An SGR ratio is 'SIGs:GNDs', without white space. SIGs and GNDs are positive integers.");
};


text_string_hint: /* empty */ { 
  yylex_hint(H_STRING); @$ = yylloc; 
  @$.first_line   = @$.last_line   = yylloc.last_line;
  @$.first_column = @$.last_column = yylloc.last_column + 1;
};

text_string:
text_string_hint TEXT_STRING {
  $$ = $2; if (!strlen($2)) _icm_msg(ERROR, &@2, "Empty text string.  Non-blank content is expected.");
  @$ = @2;
}|
text_string_hint error {
  $$ = _icm_string("");
  _icm_msg(DEBUG, &@1, "Assuming TEXT_STRING is ''.");
  _icm_msg(INFO, &@1, "A text string is basically the remainder of an input line.  It cannot be empty.");
};


text_block_hint: /* empty */ { 
  yylex_hint(H_BLOCK); @$ = yylloc; 
  @$.first_line   = @$.last_line   = yylloc.last_line;
  @$.first_column = @$.last_column = yylloc.last_column + 1;
};

text_block:
text_block_hint TEXT_BLOCK {
  $$ = $2; if (!strlen($2)) _icm_msg(ERROR, &@2, "Empty text block.  Non-blank content is expected.");
  @$ = @2;
}|
text_block_hint error {
  $$ = _icm_string("");
  _icm_msg(DEBUG, &@1, "Assuming TEXT_BLOCK is ''.");
  _icm_msg(INFO, &@1, "A text block is everything until the next keyword.  It cannot be empty.");
};


filename_hint: /* empty */ { 
  yylex_hint(H_FNAME); @$ = yylloc; 
  @$.first_line   = @$.last_line   = yylloc.last_line;
  @$.first_column = @$.last_column = yylloc.last_column + 1;
};

filename:
filename_hint FILENAME {
  $$ = $2;
  @$ = @2;
}|
filename_hint error {
  $$ = _icm_string("");
  _icm_msg(DEBUG, &@1, "Assuming FILENAME is ''.");
  _icm_msg(INFO, &@1, "A file name a sequence of lower case alpha, digits and '-', and an optional 3 character extension.");
};


real_number_hint: /* empty */ { 
  yylex_hint(H_NUMBER); @$ = yylloc; 
  @$.first_line   = @$.last_line   = yylloc.last_line;
  @$.first_column = @$.last_column = yylloc.last_column + 1;
};

real_number:
real_number_hint NUMBER { 
  char *dptr;
  double dval;
  
  /*
   * try to convert the number, ensure it completely converts.
   */
  
  dval = strtod($2, &dptr);
  if (($2 != dptr) && !dptr[0]) {  /* true if it fully converted */
    /* OK! */
  } else {
    _icm_msg(ERROR, &@2, "'%s' is not a real number.", $2);
  }

  $$ = dval;
  @$ = @2;

  _icm_msg(DEBUG, &@2, "Real conversion result: '" G_FMT "'.", $$);
}|
real_number_hint error {
  $$ = 0.0;
  _icm_msg(DEBUG, &@1, "Assuming Real NUMBER is '0.0'.");
  _icm_msg(INFO, &@1, "A real number is an integer, a real, or either followed by an exponent or standard unit.");
};


integer_hint: /* empty */ { 
  yylex_hint(H_NUMBER); @$ = yylloc; 
  @$.first_line   = @$.last_line   = yylloc.last_line;
  @$.first_column = @$.last_column = yylloc.last_column + 1;
};

integer:
integer_hint NUMBER {
  char *dptr;
  double dval = strtod($2, &dptr);
  
  /*
   * try to convert the number as a simple integer first
   * if it won't entirely convert, warn.  Then try to
   * convert to double and confirm that value is integral
   */
  
  if (($2 != dptr) && !dptr[0]) {  /* true if it fully converted */
    if (floor(dval) == ceil(dval)) {  /* true if it's an integral value */
      char *lptr;
      (void) strtol($2, &lptr, 10);
      if (($2 != lptr) && !lptr[0]) { /* true if full integer conversion */
	/* OK! */
      } else {
	_icm_msg(WARNING, &@2, "'%s' is an integer value expressed in non-integer format", $2);
      }
    } else {
      _icm_msg(ERROR, &@2, "'%s' is a non-integer value.  Converted to '%d'", $2, (long) dval);
    }
  } else {
    _icm_msg(ERROR, &@2, "'%s' is not an integer number.", $2);
  }

  $$ = (long) dval;
  @$ = @2;

  _icm_msg(DEBUG, &@2, "Integer conversion result: '%d'", $$);
}|
integer_hint error {
  $$ = 0;
  _icm_msg(DEBUG, &@1, "Assuming Integer NUMBER is '0'.");
  _icm_msg(INFO, &@1, "An integer is a whole number.  If a decimal point is present, the fractional part must be zero.");
};

/*
 * Reducing the input stream to the goal 'icm' constitutes a successful parse.
 *
 * A complete interconnect consists of (in order):
 *
 * [Begin Header]      <-- Required, FIRST.  All leading input is ignored.
 * header_part         <-- Required, in this order.
 * family_part         <-- Required, in this order.
 * section_part        <-- Required. in this order.
 * [End]               <-- Required, in order.  All trailing input is ignored.
 *
 *
 * strategy:  To process as much of the input file as possible, we actually express this as an
 *            iterator, and then check with actions and validation routines that we've 
 *            found all the right things in the correct order.
 *
 *            While is it possible to construct a grammer that rigidly enforces correctness 
 *            without all this checking, error recovery is also well-nigh impossible once 
 *            the first error is encountered.  This way we can process substantially more
 *            of the input despite errors, and generate some meaningful errors other than
 *            'Syntax error'.  (And yes, I tried this both ways, too!)
 * 
 *            There is another reason for using this tactic.  In many cases, it is quite
 *            awkward to express all of the legitimate combinations of certain ICM
 *            syntactic elements due to power of combinatorial expansion.  The combinations
 *            increase as a factorial; a couple of elements is easy, 4 or 5 difficult,
 *            more than that is an impossibly huge number.
 *
 * error recovery tactics:  
 *  Handle the three major sections of an ICM specification more or less independently.
 */


end:
END {
  _icm_msg(INFO, &@1, "%s found, finishing IBIS ICM Model compilation.", $1);
  $$ = $1;
};

icm:
icm_body end {
  $1->loc = @1;
  validate_icm($1);       /* check global self consistency */
  $$ = $1;
  _icm_emit_result($$);
  YYACCEPT;               /* Terminate parse */
}|
icm_body {
  if ($1) { 
    YYLTYPE end_loc;
    end_loc.first_line = end_loc.last_line = @1.last_line + 1;
    end_loc.first_column = end_loc.last_column = 1;
    _icm_msg(ERROR, &end_loc, "Missing required [End] Keyword.  [End] assumed.");
    $1->loc = @1;
    validate_icm($1);  /* check global self consistency */
    $$ = $1;
    _icm_emit_result($$);
    YYACCEPT;               /* Terminate parse */
  } else {
    _icm_msg(ERROR, &yylloc, "Nothing to parse.  Missing required [Begin Header] Keyword?");
    YYABORT;
  }
};

icm_body:
/* empty */ {
  $$ = (icm_t *) 0;
}|
icm_body error {
  FLUSH_KEYWORD;
  _icm_msg(INFO, &@2, "Resuming parse at next keyword.");
  yyerrok;
  $$ = $1;
}|
icm_body header_part {
  if (!$1) $1 = _icm_new_icm();
  if ($2) {
    if ($1->family_seen || $1->section_seen) _icm_msg(ERROR, &@2, "[Begin Header] must be the first keyword in the file.");
    if ($1->header) {
      _icm_msg(ERROR, &@2, "Redundant [Begin Header] or [End Header].");
      _icm_free_icm_header($2);
    } else {
      validate_icm_header($2);
      $1->header = $2;
      $$ = $1;
    }
  }
  if (!($$->header_seen || $$->family_seen || $$->section_seen)) @$ = @2;
  $$->header_seen = 1;
}|
icm_body family_part {
  if (!$1) $1 = _icm_new_icm();
  if ($2) {
    if ((!$1->header_seen) || $1->section_seen) _icm_msg(ERROR, &@2, "[Begin ICM Family] must follow [End Header].");
    if ($1->family) {
      _icm_msg(ERROR, &@2, "Redundant [Begin ICM Family] or [End ICM Family].");
      _icm_free_icm_family($2);
    } else {
      validate_icm_family($2);
      $1->family = $2;
      $$ = $1;
    }
  }
  if (!($$->header_seen || $$->family_seen || $$->section_seen)) @$ = @2;
  $$->family_seen = 1;
}|
icm_body section_part {
  if (!$1) $1 = _icm_new_icm();
  if ($2) {
    if ((!$1->header_seen) || (!$1->family_seen)) {
      _icm_msg(ERROR, &@2, "[Begin ICM Section] must follow either [End ICM Family] or a previous [End ICM Section].");
    }
    
    validate_icm_section($2);
    $1->sections = (icm_section_t **) realloc($1->sections, (1 + $1->section_count) * sizeof(icm_section_t *));
    $1->sections[$1->section_count] = $2;
    $1->section_count += 1;
    $$ = $1;
  }
  if (!($$->header_seen || $$->family_seen || $$->section_seen)) @$ = @2;
  $$->section_seen = 1;
};

/*
 * the header_part consists of:
 *
 * [ICM Ver]             <version>                     <-- MUST BE FIRST and <version> MUST be "1.0"
 * [File Name]           <file_name>                   <-- Required
 * [File Rev]            <text string(revision)>       <-- Required
 * [Date]                <text string(date, 40 char)>  <-- Optional
 * [Source]              <text block>                  <-- Optional
 * [Notes]               <text block>                  <-- Optional
 * [Disclaimer]          <text block>                  <-- Optional
 * [Copyright]           <text block>                  <-- Optional
 * [Support]             <text block>                  <-- Optional
 * [Redistribution]      <redistribution type>         <-- Required, value is 'Yes, No or Specific'
 * [Redistribution Text] <text block>                  <-- Required only if <redistribution type. is 'Specific'
 * [End Header]                                        <-- Required marks the end of all header declarations.
 *
 * error recovery tactics:
 *  Match the [End Header] keyword and try to continue.
 */

end_header:
END_HEADER newline {
  $$ = $1;
};

header_part:
header_body end_header {
  yylex_enable_x_char(); /* enable the lexical analyzer's [Comment Char] recognition feature */
  $1->loc = @$;
  $$ = $1;
}|
header_body {
  YYLTYPE end_loc;
  yylex_enable_x_char(); /* enable the lexical analyzer's [Comment Char] recognition feature */
  end_loc.first_line = end_loc.last_line = @1.last_line + 1;
  end_loc.first_column = end_loc.last_column = 1;
  _icm_msg(ERROR, &end_loc, "Missing required [End Header] keyword.  [End Header] assumed.");
  $1->loc = @$;
  $$ = $1;
};


header_body:
BEGIN_HEADER newline {
  _icm_msg(INFO, &@1, "%s found, beginning IBIS ICM Model compilation.", $1);
  $$ = _icm_new_icm_header();
}|
header_body error newline {
  yylex_enable_x_char(); /* enable the lexical analyzer's [Comment Char] recognition feature */
  _icm_msg(INFO, &@3, "Resuming parse here.");
  yyerrok;
  $$ = $1;
}|
header_body ICM_VER text_string newline {
  if ($1->version) {
    _icm_msg(ERROR, &@2, "Duplicate %s keyword.", $2);
  } else {
    if ($1->file_name || $1->file_rev || $1->date || $1->source || $1->notes || 
	$1->disclaimer || $1->copyright || $1->support || $1->redist || $1->redist_text) {
      _icm_msg(ERROR, &@2, "The [ICM Ver] keyword must be the first keyword following [Begin Header].");
    }

    /*
     * determine if this is the correct parser or not
     */

    if ($3 == _icm_string("1.0")) {      /* OK!.  That's me */
    } else {                             /* Uh Oh */
      _icm_msg(ERROR, &@3, "Unsupported version '[ICM Ver] %s'.", $3);
      $3 = _icm_string("1.0");
      _icm_msg(WARNING, &@3, "Attempting to continue as '[ICM ver] %s'.", $3);
    }

    $1->version = $3;
  }
  yylex_enable_x_char(); /* enable the lexical analyzer's [Comment Char] recognition feature */
  $$ = $1;
}|
header_body FILE_NAME filename newline { 
  if ($1->file_name) {
    _icm_msg(ERROR, &@2, "Duplicate %s keyword.", $2);
  } else {
    char *real_file_name = _icm_string(_icm_file_name());
    $1->file_name = $3;
    if ($1->file_name != real_file_name) {
      _icm_msg(WARNING, &@3, "[File Name] '%s' is not the same as the input file '%s'.", $1->file_name, real_file_name);
    }
  }
  yylex_enable_x_char(); /* enable the lexical analyzer's [Comment Char] recognition feature */
  $$ = $1;
}|
header_body FILE_REV text_string newline { 
  if ($1->file_rev) {
    _icm_msg(ERROR, &@2, "Duplicate %s keyword.", $2);
  } else {
    $1->file_rev = $3;
  }
  yylex_enable_x_char(); /* enable the lexical analyzer's [Comment Char] recognition feature */
  $$ = $1;
}|
header_body DATE text_string newline {
  if ($1->date) {
    _icm_msg(ERROR, &@2, "Duplicate %s keyword.", $2);
  } else {
    if (strlen($3) > 40) {
      _icm_msg(ERROR, &@3, "The %s keyword argument is limited to 40 characters.", $2);
    }
    $1->date = $3;
  }
  yylex_enable_x_char(); /* enable the lexical analyzer's [Comment Char] recognition feature */
  $$ = $1;
}|
header_body SOURCE text_block newline {
  if ($1->source) {
    _icm_msg(ERROR, &@2, "Duplicate %s keyword.", $2);
  } else {
    $1->source = $3;
  }
  yylex_enable_x_char(); /* enable the lexical analyzer's [Comment Char] recognition feature */
  $$ = $1;
}|
header_body NOTES text_block newline {
  if ($1->notes) {
    _icm_msg(ERROR, &@2, "Duplicate %s keyword.", $2);
  } else {
    $1->notes = $3;
  }
  yylex_enable_x_char(); /* enable the lexical analyzer's [Comment Char] recognition feature */
  $$ = $1;
}|
header_body DISCLAIMER text_block newline {
  if ($1->disclaimer) {
    _icm_msg(ERROR, &@2, "Duplicate %s keyword.", $2);
  } else {
    $1->disclaimer = $3;
  }
  yylex_enable_x_char(); /* enable the lexical analyzer's [Comment Char] recognition feature */
  $$ = $1;
}|
header_body COPYRIGHT text_block newline {
  if ($1->copyright) {
    _icm_msg(ERROR, &@2, "Duplicate %s keyword.", $2);
  } else {
    $1->copyright = $3;
  }
  yylex_enable_x_char(); /* enable the lexical analyzer's [Comment Char] recognition feature */
  $$ = $1;
}|
header_body SUPPORT text_block newline {
  if ($1->support) {
    _icm_msg(ERROR, &@2, "Duplicate %s keyword.", $2);
  } else {
    $1->support = $3;
  }
  yylex_enable_x_char(); /* enable the lexical analyzer's [Comment Char] recognition feature */
  $$ = $1;
}|
header_body REDISTRIBUTION redist_type newline {
  if ($1->redist) {
    _icm_msg(ERROR, &@2, "Duplicate %s keyword.", $2);
  } else {
    if (!$3) {
      _icm_msg(WARNING, &@3, "Assuming '[Redistribution] No'.");
      $3 = _icm_string("No");
    }
    $1->redist = $3;
  }
  yylex_enable_x_char(); /* enable the lexical analyzer's [Comment Char] recognition feature */
  $$ = $1;
}|
header_body REDISTRIBUTION_TEXT text_block newline {
  if ($1->redist_text) {
    _icm_msg(ERROR, &@2, "Duplicate %s keyword.", $2);
  } else {
    $1->redist_text = $3;
  }
  yylex_enable_x_char(); /* enable the lexical analyzer's [Comment Char] recognition feature */
  $$ = $1;
};

redist_type:
word {
  $$ = choice(&@1, $1, "Yes No Specific");
  if (!$$) {
    _icm_msg(ERROR, &@1, "'%s' is not one of 'Yes', 'No', or 'Specific'.", $1);
  }
};


/*
 * An ICM Model Family consists of the following:
 *
 * [Begin ICM Family]       <text string(family name)>     <-- Required
 * [Manufacturer]           <text string>                  <-- Required.  Order irrelevant, but before model_list
 * [ICM Family Description] <text block>                   <-- Required   Order irrelevant, but before model_list
 * model_list                                              <-- precedes models
 * models                                                  <-- follows model_list
 * pin_maps                                                <-- multiple instances, follow models, in any order
 * node_maps                                               <-- multiple instanced, follow models, in any order
 * [End ICM Family]                                        <-- Required, last.
 *
 * error recovery tactics:
 *  Handle the major sections of the Model Family as indpendently as possible.
 *  Match the [End ICM Family] keyword and try to continue
 */

end_family:
END_ICM_FAMILY newline {
  $$ = $1;
};

family_part:
family_body end_family {
  $1->loc = @$;
  $$ = $1;
}|
family_body {
  YYLTYPE end_loc;
  end_loc.first_line = end_loc.last_line = @1.last_line + 1;
  end_loc.first_column = end_loc.last_column = 1;
  _icm_msg(ERROR, &end_loc, "Missing required [End ICM Family] keyword.  [End ICM Family] assumed.");
  $1->loc = @$;
  $$ = $1;
};


family_body:
BEGIN_ICM_FAMILY text_string newline {
  $$ = _icm_new_icm_family();
  $$->name = $2;
}|
family_body error newline {
  _icm_msg(INFO, &@3, "Resuming parse here.");
  yyerrok;
  $$ = $1;
}|
family_body MANUFACTURER text_string newline {
  if ($1->manufacturer) {
    _icm_msg(ERROR, &@2, "Duplicate %s Keyword.", $2);
  } else {
    $1->manufacturer = $3;
  }
  $$ = $1;
}|
family_body ICM_FAMILY_DESCRIPTION text_block newline {
  if ($1->description) {
    _icm_msg(ERROR, &@2, "Duplicate %s Keyword.", $2);
  } else {
    $1->description = $3;
  }
  $$ = $1;
}|
family_body model_list {
  if (!$1->manufacturer) _icm_msg(ERROR, &@2, "[Manufacturer] expected before the model list.");
  if (!$1->description) _icm_msg(ERROR, &@2, "[ICM Family Description] expected before the model list.");
  if ($1->model_list_seen) {
    int i;
    _icm_msg(ERROR, &@2, "Duplicate [ICM Model List] Keyword.");
    for (i = 0; $2[i]; i++) _icm_free_icm_model($2[i]);
    free ((void *) $2);
  } else {
    for ($1->model_count = 0; $2[$1->model_count]; $1->model_count++);
    $1->models = $2;
    $1->model_list_seen = 1;
  }
  $$ = $1;
}|
family_body model {
  if (!$1->manufacturer) _icm_msg(ERROR, &@2, "[Manufacturer] expected before models.");
  if (!$1->description) _icm_msg(ERROR, &@2, "[ICM Family Description] expected before this point.");
  if (!$1->model_list_seen) _icm_msg(ERROR, &@2, "[ICM Model List] expected before this point.");
  if ($2) {
    int count;
    $1->model_seen = 1;
    for (count = 0; (count < $1->model_count) && ($2->name != $1->models[count]->name); count++);
    if (count < $1->model_count) {  /* yes, we found it in the model list */
      if ($1->models[count]->type)  /* but it's already been defined, too */ {
	_icm_msg(ERROR, &@2, "Duplicate definition of Model '%s'", $2->name);
        _icm_free_icm_model($2);
      } else {
	$2->mating = $1->models[count]->mating;
	$2->slew = $1->models[count]->slew;
	$2->image = $1->models[count]->image;
        $2->loc = @2;

        if (!$2->type) {
          _icm_msg(ERROR, &@2, "Model '%s' does not specify an ICM_model_type.", $2->name);
          $2->type = _icm_string("SLM_general");
          _icm_msg(WARNING, &@2, "Model '%s' defaulting to 'ICM_model_type %s'.", $2->name, $2->type);
        }

        _icm_free_icm_model($1->models[count]);
        $1->models[count] = $2;

	validate_icm_model($1->models[count]);  /* earliest point at which a model is complete */
      }
    } else {
      _icm_msg(WARNING, &@2, "Model '%s' was not found in the Model List.", $2->name);
      _icm_free_icm_model($2);
    }
    $$ = $1;
  }
}|
family_body pin_map {
  if (!$1->manufacturer) _icm_msg(ERROR, &@2, "[Manufacturer] expected before models.");
  if (!$1->description) _icm_msg(ERROR, &@2, "[ICM Family Description] expected before this point.");
  if (!$1->model_list_seen) _icm_msg(ERROR, &@2, "[ICM Model List] expected before this point.");
  if (!$1->model_seen) _icm_msg(ERROR, &@2, "All [Begin ICM Model] keywords expected before this point.");
  if ($2) {
    if ($2->pin_list) {

      $2->loc = @2;
      $1->pin_maps = (icm_pin_map_t **) realloc($1->pin_maps, ($1->pin_map_count + 1) * sizeof(icm_pin_map_t *));
      $1->pin_maps[$1->pin_map_count++] = $2;

      /*
       * CHECK: An Unordered [ICM Pin Map] may have neither Num_of_rows nor Num_of_columns
       */

      if (IS_UNORDERED($2)) {
	if ($2->num_of_rows > 0) {
	  _icm_msg(ERROR, &@2, "Subparameter 'Num_of_rows' disallowed by 'Pin_order Unordered'.", $2);
	}
	if ($2->num_of_columns > 0) {
	  _icm_msg(ERROR, &@2, "Subparameter 'Num_of_columns' disallowed by 'Pin_order Unordered'.", $2);
	}
      } else {
	
	/*
	 * CHECK: An Ordered Pin Map must have both Num_of_rows and Num_of_columns
	 *        and the values of each must be greater than zero
	 * CHECK: An Ordered Pin Map must have all row x col Pin_list entries
	 */
	
	if ($2->num_of_rows < 0) {
	  _icm_msg(ERROR, &@2, "Subparameter 'Num_of_rows' is required but missing.");
	  _icm_msg(WARNING, &@2, "Assuming 'Num_of_rows = 1'");
	  $2->num_of_rows = 1;
	}
	if ($2->num_of_columns < 0) {
	  _icm_msg(ERROR, &@2, "Subparameter 'Num_of_columns' is required but missing.");
	  _icm_msg(WARNING, &@2, "Assuming 'Num_of_columns = 1'");
	  $2->num_of_columns = 1;
	}
	
	if (($2->num_of_rows > 0) && ($2->num_of_columns > 0) &&
	    ($2->pin_list_count != ($2->num_of_rows * $2->num_of_columns))) {
	  _icm_msg(ERROR, &@2, "Pin List '%s' (%d by %d) requires %d entries, but the list has %d.", 
		   $2->name, $2->num_of_rows, $2->num_of_columns, $2->num_of_rows * $2->num_of_columns, $2->pin_list_count);
	}
      }
    } else {
      _icm_msg(ERROR, &@2, "[ICM Pin Map] '%s' lacks the required Pin_list.", $2->name);
    }
    $$ = $1;
  }
}|
family_body node_map {
  if (!$1->manufacturer) _icm_msg(ERROR, &@2, "[Manufacturer] expected before models.");
  if (!$1->description) _icm_msg(ERROR, &@2, "[ICM Family Description] expected before this point.");
  if (!$1->model_list_seen) _icm_msg(ERROR, &@2, "[ICM Model List] expected before this point.");
  if (!$1->model_seen) _icm_msg(ERROR, &@2, "All [Begin ICM Model] keywords expected before this point.");
  if ($2) {
    $2->loc = @2;
    $1->node_maps = (icm_node_map_t **) realloc($1->node_maps, ($1->node_map_count + 1) * sizeof(icm_node_map_t *));
    $1->node_maps[$1->node_map_count++] = $2;
    $$ = $1;
  }
};


/*
 * model_list consists of the following:
 *
 * [ICM Model List]                                <-- required keyword
 * <model name> <mating> <slew> <image>            <-- One or more lines composed as this one, where:
 *                                                     mating is 'Mated, Unmated_side_A, or Unmated_side_B'
 *                                                     slew is a real number
 *                                                     image is an optional file name containing a picture.
 *
 * Every model named in the list must be subsequently referenced by a [Begin ICM Model] <model name>
 *
 * error recovery strategy
 *  keep processing list entries as long as we see leading symbolic names
 *
 */

model_list:
ICM_MODEL_LIST newline {
  yylex_disable_subp(); /* no further subparameters are allowed for this keyword */
  $$ = (icm_model_t **) calloc(1, (0 + 1) * sizeof(icm_model_t *));
}|
model_list model_entry {
  if ($2) {
    int count;
    for (count = 0; $1[count]; count++) {
      if ($2->name == $1[count]->name) {
	_icm_msg(ERROR, &@2, "Duplicate Model Name '%s'", $2->name);
        _icm_free_icm_model($2);  /* free duplicate entry */
        break;
      }
    }
    if (!$1[count]) {
      $1[count] = $2;
      $1[count]->loc = @2;  /* location of list entry, but will be replaced by that of model if it is defined */
      count += 1;
      $1 = (icm_model_t **) realloc($1, (count + 1) * sizeof(icm_model_t *));
      $1[count] = (icm_model_t *) 0;
    }
    $$ = $1;
  }
};

model_entry:
word mating slew filename_hint image newline {
  $$ = _icm_new_icm_model();
  $$->name = $1;
  if (!$2) {
    _icm_msg(WARNING, &@2, "Assuming 'Mated'.");
    $2 = _icm_string("Mated");
  }
  $$->mating = $2;
  $$->slew = $3;
  $$->image = $5;
}|
error newline { 
  _icm_msg(INFO, &@2, "Resuming parse here.");
  yyerrok;
  $$ = (icm_model_t *) 0;
};

mating:
word {
  $$ = choice(&@1, $1, "Mated Unmated_side_A Unmated_side_B");
  if (!$$) {
    _icm_msg(ERROR, &@1, "'%s' is not one of: 'Mated', 'Unmated_side_A', or 'Unmated_side_B'.", $1);
  }
};

slew:
real_number {
  if ($1 < 0) _icm_msg(WARNING, &@1, "Questionable negative value '" G_FMT "'.  Slew rates should be positive.", $1);
  $$ = $1;
};

image:
/* empty */ {
  $$ = (char *) NULL;
}|
filename {
  switch (_icm_file_exists($1)) {
  case -1:
    _icm_msg(WARNING, &@1, "File '%s' does not exist in the current working directory.", $1);
    break;
  case 0:
    break;
  case 1:
    _icm_msg(WARNING, &@1, "File '%s' does not exist within directory '%s'.", $1, _icm_dir_name());
    break;
  default:
    _icm_msg(ERROR, &@1, "FIXME: Internal fault, impossible file existance check result.");
  }
  $$ = $1;
};

/*
 * a model is:
 *
 * [ICM Begin Model] <model name>         <-- Required.  <model name> must be in the model list
 * ICM_model_type <model type>            <-- Required next.  'SLM_general, SLM_quiescent, SLM_even_mode, SLM_odd_mode, MLM or S-parameter
 * SGR             <INTEGER>:<INTEGER>    <-- Optional, unless <model type> is SLM_general
 * Ref_impedance = <INTEGER>              <-- Optional, default 50.  Omited if <model type> is S-parameter [enforce this?] 
 * [ICM Model Description]  <text block>  <-- Optional
 * [Tree Path Description]  tree_body     <-- One of Tree/Nodal path required, not both
 * [Nodal Path Description] nodal_body    <-- One of Tree/Nodal path required, not both
 * [ICM Swath Parameters]                 <-- Optional
 * Left_edge = <INTEGER>                  <-- required if keyword if present, any order
 * Right_edge = <INTEGER>                 <-- required if keyword if present, any order
 * Top_edge = <INTEGER>                   <-- required if keyword if present, any order
 * Bottom_edge = <INTEGER>                <-- required if keyword if present, any order
 * [ICM Swath Pin Numbers] swath_pin_list <-- Required if swath parameters present, must be absent otherwise
 * [End ICM Model]                        <-- Required.
 *
 *
 * error recovery tactics:
 *  look for out of order parts.
 *  look for missing parts until it's obvious we won't find any.
 *  If all else fails, try to find the [End ICM Model]
 */

end_model:
END_ICM_MODEL newline {
  $$ = $1;
};

model:
model_body end_model {
  $1->loc = @$;
  $$ = $1;
}|
model_body {
  YYLTYPE end_loc;
  end_loc.first_line = end_loc.last_line = @1.last_line + 1;
  end_loc.first_column = end_loc.last_column = 1;
  _icm_msg(ERROR, &end_loc, "Missing required [End ICM Model] keyword.  [End ICM Model] assumed.");
  $1->loc = @$;
  $$ = $1;
};

model_body:
BEGIN_ICM_MODEL word newline {
  $$ = _icm_new_icm_model();
  $$->name = $2;
  $$->left_edge = -1;  /* this is a illegal value to determine if the subparameter was used */
  $$->right_edge = -1;
  $$->top_edge = -1;
  $$->bottom_edge = -1;
}|
BEGIN_ICM_MODEL word error newline {
  $$ = _icm_new_icm_model();
  $$->name = $2;
  $$->left_edge = -1;  /* this is a illegal value to determine if the subparameter was used */
  $$->right_edge = -1;
  $$->top_edge = -1;
  $$->bottom_edge = -1;
}|
model_body error newline {
  _icm_msg(INFO, &@3, "Resuming parse here.");
  yyerrok;
  $$ = $1;
}|
model_body ICM_model_type model_type newline {
  if ($1->sgr || $1->ref_impedance) _icm_msg(ERROR, &@2, "Subparameter '%s' must immediately follow [Begin ICM Model]", $2);
  if ($1->type) {
    _icm_msg(ERROR, &@2, "Duplicate '%s' subparameter.", $2);
  } else {
    if (!$3) {
      $3 = _icm_string("SLM_general");
      _icm_msg(WARNING, &@2, "Model '%s' defaulting to 'ICM_model_type %s'.", $1->name, $3);
    }
    $1->type = $3;
  }
  $$ = $1;
}|
model_body SGR sgr_ratio newline {
  if (!$1->type) _icm_msg(ERROR, &@2, "ICM_model_type expected before this point.");
  if ($1->sgr) {
    _icm_msg(ERROR, &@2, "Duplicate '%s' subparameter.", $2);
  } else {
    $1->sgr = $2;
  }
  $$ = $1;
}|
model_body Ref_impedance_EQ real_number newline {
  if (!$1->type) _icm_msg(ERROR, &@2, "ICM_model_type expected before this point.");
  if ($1->ref_impedance) {
    _icm_msg(ERROR, &@2, "Duplicate '%s' subparameter.", $2);
  } else {
    $1->ref_impedance = $3;
  }
  $$ = $1;
}|
model_body ICM_MODEL_DESCRIPTION text_block newline {
  if (!$1->type) _icm_msg(ERROR, &@2, "ICM_model_type expected before this point.");
  if ($1->description) {
    _icm_msg(ERROR, &@2, "Duplicate %s keyword.", $2);
  } else {
    $1->description = $3;
  }
  $$ = $1;
}|
model_body TREE_PATH_DESCRIPTION newline tree_body {
  if (!$1->type) {
    _icm_msg(ERROR, &@2, "ICM_model_type expected before this point.");
    /*
     * establish a default value here, code below refers to it
     */
    $1->type = _icm_string("SLM_general");
    _icm_msg(WARNING, &@2, "Model '%s' defaulting to 'ICM_model_type %s'.", $1->name, $1->type);
  } 
  if (_icm_string("S-parameter") == $1->type) {
    _icm_msg(ERROR, &@2, "S-parameter models may not use a [Tree Path Description]");
    _icm_free_icm_tree_path($4);
  } else if ($1->node_path) {
    _icm_msg(ERROR, &@2, "%s keyword excluded due to presence of [Nodal Path Description].", $2);
    _icm_free_icm_tree_path($4);
  } else if ($1->tree_path) {
    _icm_msg(ERROR, &@2, "Duplicate %s keyword.", $2);
    _icm_free_icm_tree_path($4);
  } else {
    icm_tree_path_t *link = $4;
    /*
     * CHECK: the path must start and end with a Model_pinmap
     * CHECK: the path must include at least one Section
     */
    if (link) {
      int have_section = 0;
      if (!IS_PINMAP(link)) _icm_msg(ERROR, &@4, "%s must begin with a Model_pinmap.", $2);
      while (link->next) {
	if (IS_SECTION(link)) have_section =1;
	link = link->next;
      }
      if (!have_section) _icm_msg(ERROR, &@4, "%s must have at least one Section.", $2);
      if (!IS_PINMAP(link)) {
	YYLTYPE loc;
	loc.first_line = loc.last_line = @4.last_line;
	loc.first_column = loc.last_column = 1;
	_icm_msg(ERROR, &loc, "%s must end with a Model_pinmap.", $2);
      }
    } else {
      _icm_msg(ERROR, &@4, "%s is empty.  There must be a pinmap, at least one section, and a pinmap.", $2);
    }
    $1->tree_path = $4;
  }
  $$ = $1;
}|
model_body NODAL_PATH_DESCRIPTION newline nodal_body {
  if (!$1->type) _icm_msg(ERROR, &@2, "ICM_model_type expected before this point.");
  if ($1->swath_present) {
    int i;
    _icm_msg(ERROR, &@2, "%s is disallowed due to the use of [ICM Swath Parameters].", $2);
    for (i = 0; $4[i]; i++) _icm_free_icm_node_path($4[i]);
    free((void *)$4);
  } else if ($1->tree_path) {
    int i;
    _icm_msg(ERROR, &@2, "%s keyword excluded due to presence of [Tree Path Description].", $2);
    for (i = 0; $4[i]; i++) _icm_free_icm_node_path($4[i]);
    free((void *)$4);
  } else if ($1->node_path) {
    int i;
    _icm_msg(ERROR, &@2, "Duplicate %s keyword.", $2);
    for (i = 0; $4[i]; i++) _icm_free_icm_node_path($4[i]);
    free((void *)$4);
  } else {
    /* 
     * CHECK: Ensure there is AT LEAST one n_section and one nodemap in the node path
     */
    int n_section_count = 0;
    int nodemap_count = 0;

    $1->node_path = $4;

    for ($1->node_path_count = 0; $4[$1->node_path_count]; $1->node_path_count++) {
      if (IS_NODEMAP($4[$1->node_path_count])) nodemap_count +=1;
      if (IS_N_SECTION($4[$1->node_path_count])) n_section_count +=1;
    }
    if (nodemap_count < 1) _icm_msg(ERROR, &@4, "[Nodal Path Description] requires at least one Model_nodemap.");
    if (n_section_count < 1) _icm_msg(ERROR, &@4, "[Nodal Path Description] requires at least one N_section.");
  }
  $$ = $1;
}|
model_body ICM_SWATH_PARAMETERS newline {
  if (!$1->type) _icm_msg(ERROR, &@2, "ICM_model_type expected before this point.");
  if ($1->node_path) {
    _icm_msg(ERROR, &@2, "%s is disallowed due to the use of [Nodal Path Description].", $2);
  } else {
    if ($1->swath_present) {
      _icm_msg(ERROR, &@2, "Duplicate %s keyword.",$2);
    } else {
      $1->swath_present = 1;
    }
  }
  $$ = $1;
}|
model_body Left_edge_EQ integer newline { 
  if ($1->swath_present) {
    if ($1->left_edge >= 0) {
      _icm_msg(ERROR, &@2, "Duplicate '%s' subparameter.", $2);
    } else if ($3 < 0) {
      _icm_msg(ERROR, &@3, "Illegal negative value '%d'. '%s' must be non-negative.", $3, $2);
      _icm_msg(WARNING, &@3, "Assuming '%s = 0'.", $2);
      $3 = 0;
    }
    $1->left_edge = $3;
  }
  $$ = $1;
}|
model_body Right_edge_EQ integer newline { 
  if ($1->swath_present) {
    if ($1->right_edge >= 0) {
      _icm_msg(ERROR, &@2, "Duplicate '%s' subparameter.", $2);
    } else if ($3 < 0) {
      _icm_msg(ERROR, &@3, "Illegal negative value '%d'. '%s' must be non-negative.", $3, $2);
      _icm_msg(WARNING, &@3, "Assuming '%s = 0'.", $2);
      $3 = 0;
    }
    $1->right_edge = $3;
  }
  $$ = $1;
}|
model_body Top_edge_EQ integer newline { 
  if ($1->swath_present) {
    if ($1->top_edge >= 0) {
      _icm_msg(ERROR, &@2, "Duplicate '%s' subparameter.", $2);
    } else if ($3 < 0) {
      _icm_msg(ERROR, &@3, "Illegal negative value '%d'. '%s' must be non-negative.", $3, $2);
      _icm_msg(WARNING, &@3, "Assuming '%s = 0'.", $2);
      $3 = 0;
    }
    $1->top_edge = $3;
  }
  $$ = $1;
}|
model_body Bottom_edge_EQ integer newline { 
  if ($1->swath_present) {
    if ($1->bottom_edge >= 0) {
      _icm_msg(ERROR, &@2, "Duplicate '%s' subparameter.", $2);
    } else if ($3 < 0) {
      _icm_msg(ERROR, &@3, "Illegal negative value '%d'. '%s' must be non-negative.", $3, $2);
      _icm_msg(WARNING, &@3, "Assuming '%s = 0'.", $2);
      $3 = 0;
    }
    $1->bottom_edge = $3;
  }
  $$ = $1;
}|
model_body ICM_SWATH_PIN_NUMBERS newline { yylex_disable_subp(); /* no more subparameters for keyword */ } swath_pin_list {
  if (!$1->type) _icm_msg(ERROR, &@2, "ICM_model_type expected before this point.");
  if ($1->node_path) {
    _icm_msg(ERROR, &@2, "%s is disallowed due to the use of [Nodal Path Description].", $2);
  } else {
    if ($1->swath) {
      int i;
      _icm_msg(ERROR, &@2, "Duplicate %s keyword.", $2);
      for (i = 0; $5[i]; i++) _icm_free_icm_swath_entry($5[i]);
      free((void *)$5);
    } else {
      icm_swath_entry_t **pin;
      
      /*
       * count the entries and evaluate the physical dimensions of the swath
       * ensure there are enough entries.
       */
      
      $1->swath = $5;
      $1->swath_rows = 0;
      $1->swath_cols = 0;
      $1->swath_count = 0;
      
      for (pin = $1->swath; pin && *pin; pin++) {
	$1->swath_count += 1;
	if ((*pin)->row > $1->swath_rows) $1->swath_rows = (*pin)->row;
	if ((*pin)->col > $1->swath_cols) $1->swath_cols = (*pin)->col;
      }
      
      if ($1->swath_count != ($1->swath_rows * $1->swath_cols)) {
	_icm_msg(ERROR, &@5, "The swath pin array is %d by %d [%d elements], but there are %d swath pin entries.", 
		 $1->swath_rows, $1->swath_cols, $1->swath_rows * $1->swath_cols, $1->swath_count);
      }
      
      /* 
       * CHECK: we have all ROW x COL swath pin entries without any duplication or omission.
       */
      
      pin = $1->swath;
      if (pin) {
	int r, c;

	_icm_msg(DEBUG, &@5, "Examining %d Swath entries organized %d by %d.",
		 $1->swath_count, $1->swath_rows, $1->swath_cols);

	for (r = 1; r <= $1->swath_rows; r++) {
	  for (c = 1; c <= $1->swath_cols; c++) {
	    int n, i;
	    n = 0;

	    for (i = 0; i < $1->swath_count; i++) {
	      if ((pin[i]->row == r) && (pin[i]->col == c)) n += 1;
	    }
	    if (1 > n) _icm_msg(ERROR, &@5, "Swath pin entry (%d, %d) is missing.", r, c);
	    if (1 < n) _icm_msg(ERROR, &@5, "Swath pin entry (%d, %d) is duplicated.", r, c);
	    if (1 == n) _icm_msg(DEBUG, &@5, "Swath pin entry (%d, %d) is just right.", r, c);
	  }
	}
      }
      /*
       * We cannot check the swath list against the swath parameters at this time.
       * These two keywords may appear in either order.
       * The validate_model routine will have to deal with it when the model is complete.
       */
    }
  }
  $$ = $1;
};

model_type:
word {
  $$ = choice(&@1, $1, "SLM_general SLM_quiescent SLM_even_mode SLM_odd_mode MLM S-parameter");
  if (!$$) {
    _icm_msg(ERROR, &@1, "'%s' is not one of: 'SLM_general', 'SLM_quiescent', 'SLM_even_mode', 'SLM_odd_mode', 'MLM', or 'S-parameter'.", $1);
  }
};


/*
 * A tree consists of a sequence of one or more sections 
 * and zero or more forks capped on each end by a pinmap.
 *
 * A fork has zero or more sections and an optional terminating pinmap.
 *
 * Forks may include other forks without limit.
 *
 * The result is simply a binary tree.
 *
 * Elements are defined as following:
 *
 * model_pinmap:   Model_pinmap <pinmap name>                       <-- the <pinmap name> must be defined after [End ICM Model].
 * model_section:  Section {Mult, Len} = <value> <section name>     <-- the <section name> must be defined after [End IDM Family].
 *                                                                      <value> is integer for Mult, real for Len
 * Fork
 * zero or more sections, followed by an optional pinmap.
 * Endfork
 * 
 * error recovery tactics:
 *  Collect the various elements in a very general way, 
 *  otherwise errors cause large parts of the path to be skipped.
 *  Enforce relationships only when reducing significant structures.
 */

tree_body: 
model_pinmap { 
  $$ = $1;
}|
interconnect {
  $$ = $1;
}|
tree_body error newline {
  _icm_msg(INFO, &@3, "Resuming parse here.");
  yyerrok;
  $$ = $1;
}|
tree_body model_pinmap {
  icm_tree_path_t *link = $1;
  if (link) {
    while (link->next) link = link->next;
    link->next = $2;
  } else {
    $1 = $2;
  }
  $$ = $1;
}|
tree_body interconnect {
  icm_tree_path_t *link = $1;
  if (link) {
    while (link->next) link = link->next;
    link->next = $2;
  } else {
    $1 = $2;
  }
  $$ = $1;
};

model_pinmap:
Model_pinmap word newline {
  $$ = _icm_new_icm_tree_path();
  $$->type = $1;
  $$->name = $2;
  $$->loc = @$;
};

interconnect:
model_section {
  $$ = $1;
}|
fork {
  $$ = $1;
}|
interconnect error newline {
  _icm_msg(INFO, &@3, "Resuming parse here.");
  yyerrok;
  $$ = $1;
}|
interconnect model_section {
  icm_tree_path_t *link = $1;
  if (link) {
    while (link->next) link = link->next;
    link->next = $2;
  } else {
    $1 = $2;
  }
  $$ = $1;
}| 
interconnect fork {
  icm_tree_path_t *link = $1;
  if (link) {
    while (link->next) link = link->next;
    link->next = $2;
  } else {
    $1 = $2;
  }
  $$ = $1;
};

model_section:
Section Mult_EQ integer word newline { 
  $$ = _icm_new_icm_tree_path();
  $$->type = $1;
  $$->scaletype = $2;
  if ($3 <= 0) {
    _icm_msg(ERROR, &@2, "'Mult = %d' must use a positive, non-zero integer", $3);
    _icm_msg(WARNING, &@2, "Assuming 'Mult = 1'.");
    $3 = 1;
  }
  $$->scalefactor = $3;
  $$->name = $4;
  $$->loc = @$;
}|
Section Len_EQ real_number word newline { 
  $$ = _icm_new_icm_tree_path();
  $$->type = $1;
  $$->scaletype = $2;
  if ($3 <= 0) {
    _icm_msg(ERROR, &@2, "'Len = " G_FMT "' must use a positive, non-zero real number", $3);
    _icm_msg(WARNING, &@2, "Assuming 'Len = 1.0'.");
    $3 = 1.0;
  }
  $$->scalefactor = $3;
  $$->name = $4;
  $$->loc = @$;
}|
Section error newline { 
  _icm_msg(INFO, &@1, "Hint: 'Section {Len|Mult}=<number> <section-name>'");
  yyerrok;
  $$= (icm_tree_path_t *) 0;
};


endfork:
Endfork newline {
};

fork:
fork_body endfork {
  icm_tree_path_t *link = $1->fork;
  if (link) {
    int pinmap_count = 0;
    while(link->next) {
      if (IS_PINMAP(link)) {
	pinmap_count += 1;
	if (link->next) {
	  _icm_msg(ERROR, &link->loc, "Pin Map '%s' must be the last member of the fork.", link->name);
	}
      }
      link = link->next;
    }
    if (pinmap_count > 1) {
      _icm_msg(ERROR, &@1, "A fork may have at most one Pin Map.");
    }
  }
  $1->loc = @$;
  $$ = $1;
}|
fork_body {
  icm_tree_path_t *link = $1->fork;
  YYLTYPE end_loc;
  end_loc.first_line = end_loc.last_line = @1.last_line;
  end_loc.first_column = end_loc.last_column = 1;
  _icm_msg(ERROR, &end_loc, "Missing required Endfork subparameter.  Endfork assumed.");
  if (link) {
    int pinmap_count = 0;
    while(link->next) {
      if (IS_PINMAP(link)) {
	pinmap_count += 1;
	if (link->next) {
	  _icm_msg(ERROR, &link->loc, "Pin Map '%s' must be the last member of the fork.", link->name);
	}
      }
      link = link->next;
    }
    if (pinmap_count > 1) {
      _icm_msg(ERROR, &@1, "A fork may have at most one Pin Map.");
    }
  }
  $1->loc = @$;
  $$ = $1;
};

fork_body: 
Fork newline { 
  $$ = _icm_new_icm_tree_path();
  $$->type = $1;
}|
fork_body error newline { 
  _icm_msg(INFO, &@3, "Resuming parse at this point.");
  yyerrok;
  $$ = $1;
}|
fork_body model_pinmap { 
  icm_tree_path_t *link = $1->fork;
  if (link) {
    while(link->next) link = link->next;
    link->next = $2;
  } else {
    $1->fork = $2;
  }
  $$ = $1;
}|
fork_body interconnect { 
  icm_tree_path_t *link = $1->fork;
  if (link) {
    while(link->next) link = link->next;
    link->next = $2;
  } else {
    $1->fork = $2;
  }
  $$ = $1;
};

/*
 * nodal_body
 *
 * consists of at least one nodemap and at least one n_section in any order and quantity.
 *
 * Model_Nodemap <nodemap name>                              <-- <nodemap name> must be defined later in ICM section
 * N_section ( <node name> ... ) scale_factor <name>         <-- parenthesized list of <node name> spanning any number of lines
 *
 * error recovery tactics:
 *  Try the next line...
 */

nodal_body: 
model_nodemap
{ 
  $$ = (icm_node_path_t **) calloc(1, (1 + 1 ) * sizeof(icm_node_path_t *));
  $$[0] = $1;
  $$[1] = (icm_node_path_t *) 0;
}| 
n_section
{ 
  $$ = (icm_node_path_t **) calloc(1, (1 + 1 ) * sizeof(icm_node_path_t *));
  $$[0] = $1;
  $$[1] = (icm_node_path_t *) 0;
}|
nodal_body error newline {
  _icm_msg(INFO, &@3, "Resuming parse here.");
  yyerrok;
  $$ = $1;
}| 
nodal_body n_section
{
  int count;
  for (count = 0; $1[count]; count++);
  $1[count] = $2;
  count += 1;
  $1 = (icm_node_path_t **) realloc($1, (count + 1) * sizeof(icm_node_path_t *));
  $1[count] = (icm_node_path_t *) 0;
  $$ = $1;
}|
nodal_body model_nodemap
{ 
  int count;
  for (count = 0; $1[count]; count++);
  $1[count] = $2;
  count += 1;
  $1 = (icm_node_path_t **) realloc($1, (count + 1) * sizeof(icm_node_path_t *));
  $1[count] = (icm_node_path_t *) 0;
  $$ = $1;
};

model_nodemap: 
Model_nodemap word newline { 
  $$ = _icm_new_icm_node_path();
  $$->type = $1;
  $$->name = $2;
  $$->loc = @$;
};

n_section: 
N_section node_list Mult_EQ integer word newline { 
  $$ = _icm_new_icm_node_path();
  $$->type = $1;
  for($$->node_count = 0; $2 && $2[$$->node_count]; $$->node_count++);
  $$->nodes = $2;
  $$->scaletype = $3;
  if ($4 <= 0.0) {
    _icm_msg(ERROR, &@3, "'Mult = %d' must use a positive, non-zero integer", $4);
    _icm_msg(WARNING, &@3, "Assuming 'Mult = 1'.");
    $4 = 1;
  }
  $$->scalefactor = $4;
  $$->name = $5;
  $$->loc = @$;
}|
N_section node_list Len_EQ real_number word newline { 
  $$ = _icm_new_icm_node_path();
  $$->type = $1;
  for($$->node_count = 0; $2 && $2[$$->node_count]; $$->node_count++);
  $$->nodes = $2;
  $$->scaletype = $3;
  if ($4 <= 0.0) {
    _icm_msg(ERROR, &@3, "'Len = " G_FMT "' must use a positive, non-zero real number", $4);
    _icm_msg(WARNING, &@3, "Assuming 'Len = 1.0'.");
    $4 = 1.0;
  }
  $$->scalefactor = $4;
  $$->name = $5;
  $$->loc = @$;
}|
N_section error newline { 
  _icm_msg(INFO, &@1, "Hint: 'N_section ( list of names ) {Len|Mult}=<number> <section-name>'");
  yyerrok;
  $$= (icm_node_path_t *) 0;
};

node_list: 
'(' list_of_names ')' {
  $$ = $2;
};

list_of_names: 
newline node_name_hint {
  $$ = (char **) 0;
}| 
node_name {
  int size = 0;
  $$ = (char **) calloc(1, (size + 2) * sizeof(char *));
  $$[size] = $1;
  $$[size+1] = (char *) 0;
}| 
list_of_names newline node_name_hint { 
  $$ = $1;
}|
list_of_names node_name { 
  int size = 0;
  for(size = 0; $1 && $1[size]; size++);
  $$ = (char **) realloc($1, (size + 2) * sizeof(char *));
  $$[size] = $2;
  $$[size+1] = (char *) 0;
  /* 
   * CHECK: node names may not exceed 20 characters
   */
  if (strlen($2) > 20) _icm_msg(ERROR, &@2, "Node Name '%s' exceeds the maximum of 20 characters.", $2);
};


/*
 * swath_pin_list
 *
 * consists of a series of one or more swath_pin_entry
 *
 * swath_pin_entry
 *
 * <NAME> <INTEGER> <INTEGER>
 *
 * error recovery tactics:
 *  keep trying as long as we see at least a leading name.
 *  give up on anything else
 *
 */

swath_pin_list: 
swath_pin_entry { 
  int count = 0;
  $$ = (icm_swath_entry_t **) calloc(1, (count + 2) * sizeof(icm_swath_entry_t *));
  $$[count++] = $1;
  $$[count] = (icm_swath_entry_t *) 0;
}| 
swath_pin_list swath_pin_entry { 
  if ($2) {
    int count;
    for (count = 0; $1[count]; count++);
    $1 = (icm_swath_entry_t **) realloc($1, (count + 2) * sizeof(icm_swath_entry_t *));
    $1[count++] = $2;
    $1[count] = (icm_swath_entry_t *) 0;
  }
  $$ = $1;
};

swath_pin_entry: 
word integer integer newline { 
  $$ = _icm_new_icm_swath_entry();
  $$->name = $1;
  if ($2 <= 0) {
    _icm_msg(ERROR, &@2, "Swath List Physical Row '%d' must be greater than 0.", $2);
    _icm_msg(WARNING, &@2, "Assuming Row '1'.");
    $2 = 1;
  }
  $$->row = $2;
  if ($3 <= 0) {
    _icm_msg(ERROR, &@3, "Swath List Physical Column '%d' must be greater than 0.", $3);
    _icm_msg(WARNING, &@3, "Assuming Column '1'.", $3);
    $3 = 1;
  }
  $$->col = $3;
}|
error newline {
  _icm_msg(INFO, &@2, "Resuming parse here");
  yyerrok;
  $$ = (icm_swath_entry_t *) 0;
};

/*
 * [ICM Pin Map]
 *
 * consists of:
 *
 * [ICM Pin Map] <map_name>                                  <-- Required if <map_name> previously referenced.
 * Pin_order {Row_ordered, Column_ordered, Unordered}        <-- required subparameter.  FIRST
 * Num_of_rows <INTEGER>                                     <-- omitted if unordered.   SECOND
 * Num_of_columns <INTEGER>                                  <-- omitted if unordered.   THIRD
 * Pin_list                                                  <-- marks beginning of the pinlist.  LAST
 * <WORD> <WORD>                                             <-- repeated until next [keyword]
 *
 * error recovery tactics:
 *  The Lexical analyzer must report Num_of_columns and Num_of_rows as symbolic names if
 *  their use is forbidden by the Pin_order value 'Unordered'.  
 *  We use a control point to force such behavior in that particular case.
 *  This is done to prevent polluting the user name space should the user actually want
 *  to use these names for some purpose, hard as it is to imagine.
 */

pin_map: 
ICM_PIN_MAP word newline { 
  $$ = _icm_new_icm_pin_map();
  $$->num_of_rows = -1;
  $$->num_of_columns = -1;
  $$->name = $2;
}|
ICM_PIN_MAP word error newline { 
  $$ = _icm_new_icm_pin_map();
  $$->num_of_rows = -1;
  $$->num_of_columns = -1;
  $$->name = $2;
}|
pin_map error newline {
  _icm_msg(INFO, &@3, "Resuming parse here");
  yyerrok;
  $$ = $1;
}|
pin_map pin_order newline {
  if ($1->pin_order) _icm_msg(ERROR, &@2, "Redundant subparameter '%s'.", $2);
  if (!$2) {
    _icm_msg(WARNING, &@2, "Assuming 'Pin_order Unordered'.");
    $2 = _icm_string("Unordered");
  }
  $1->pin_order = $2;
  $$ = $1;
}|
pin_map Num_of_rows_EQ integer newline { 
  if (!$1->pin_order) _icm_msg(ERROR, &@2, "Expected subparameter Pin_order before this point.");
  if ($1->num_of_rows > 0) {
    _icm_msg(ERROR, &@2, "Redundant subparameter '%s'.", $2);
  } else {
    if ($3 < 0) {
      _icm_msg(ERROR, &@3, "Subparameter '%s' requires a positive, non-zero integer value.", $2);
      _icm_msg(WARNING, &@3, "Assuming '%s = 1'", $2);
      $3 = 1;
    }
    $1->num_of_rows = $3;
  }
  $$ = $1;
}| 
pin_map Num_of_columns_EQ integer newline { 
  if (!$1->pin_order) {
    _icm_msg(ERROR, &@2, "Expected subparameter Pin_order before this point.");
    _icm_msg(WARNING, &@2, "Assuming 'Pin_order Unordered'.");
    $1->pin_order = _icm_string("Unordered");
  }
  if (IS_ORDERED($1) && $1->num_of_rows < 0) _icm_msg(ERROR, &@2, "Expected subparameter Num_of_rows before this point.");
  if ($1->num_of_columns > 0) {
    _icm_msg(ERROR, &@2, "Redundant subparameter '%s'.", $2);
  } else {
    if ($3 < 0) {
      _icm_msg(ERROR, &@3, "Subparameter '%s' requires a positive, non-zero integer value.", $2);
      _icm_msg(WARNING, &@3, "Assuming '%s = 1'", $2);
      $3 = 1;
    }
    $1->num_of_columns = $3;
  }
  $$ = $1;
}|
pin_map Pin_list newline { yylex_disable_subp(); /* no more subparameters */ } pin_list { 
  if (!$1->pin_order) {
    _icm_msg(ERROR, &@2, "Expected subparameter Pin_order before this point.");
    _icm_msg(WARNING, &@2, "Assuming 'Pin_order Unordered'.");
    $1->pin_order = _icm_string("Unordered");
  }
  if (IS_ORDERED($1)) {
    if ($1->num_of_rows < 0) _icm_msg(ERROR, &@2, "Expected subparameter Num_of_rows before this point.");
    if ($1->num_of_columns < 0) _icm_msg(ERROR, &@2, "Expected subparameter Num_of_columns before this point.");
  }
  if ($1->pin_list) {
    int i;
    _icm_msg(ERROR, &@2, "Redundant subparameter '%s'.", $2);
    for (i = 0; $5[i]; i++) _icm_free_icm_pin_list_entry($5[i]);
    free((void *)$5);
  } else {
    for ($1->pin_list_count = 0; $5[$1->pin_list_count]; $1->pin_list_count++);
    $1->pin_list = $5;
  }
  $$ = $1;
};

pin_order:
Pin_order word { 
  $$ = choice(&@2, $2, "Row_ordered Column_ordered Unordered");
  if (!$$) {
    _icm_msg(ERROR, &@2, "'%s' is not one of: 'Row_ordered', 'Column_ordered', or 'Unordered'.", $2);
  }
};

pin_list:
pin_list_entry {
  $$ = (icm_pin_list_entry_t **) calloc(1, (0 + 2) * sizeof(icm_pin_list_entry_t *));
  $$[0] = $1;
  $$[1] = (icm_pin_list_entry_t *) 0;
}|
pin_list pin_list_entry {
  int count;
  for (count = 0; $1[count]; count++);
  $1 = (icm_pin_list_entry_t **) realloc($1, (count + 2) * sizeof(icm_pin_list_entry_t *));
  $1[count++] = $2;
  $1[count] = (icm_pin_list_entry_t *)0;
  $$ = $1;
};

pin_list_entry:
word word newline {
  $$ = _icm_new_icm_pin_list_entry();
  $$->pinname = $1;
  $$->bookname = $2;
}|
error newline {
  _icm_msg(INFO, &@2, "Resuming parse here");
  yyerrok;
  $$ = (icm_pin_list_entry_t *) 0;;
};

/*
 * node_map consists of
 *
 * [ICM Node Map] <nodemap name>              <-- Required if <nodemap name> was previously referenced by a model
 * pin node name                              <-- <word> <name> <word>  repeated until the next keyword
 *
 * error recovery tactics:
 *  keep reading as long as we see at least the first integer
 */


node_map: 
ICM_NODE_MAP word newline { 
  $$ = _icm_new_icm_node_map();
  $$->name = $2;
}|
ICM_NODE_MAP word error newline { 
  $$ = _icm_new_icm_node_map();
  $$->name = $2;
}|
node_map error newline {
  _icm_msg(INFO, &@3, "Resuming parse here.");
  yyerrok;
  $$ = $1;
}|
node_map word node_name word newline { 
  $1->node_list = (icm_node_list_entry_t **) realloc($1->node_list, ($1->node_list_count + 2) * sizeof(icm_node_list_entry_t *));
  $1->node_list[$1->node_list_count] = _icm_new_icm_node_list_entry();
  $1->node_list[$1->node_list_count]->pinname = $2;
  $1->node_list[$1->node_list_count]->nodename = $3;
  $1->node_list[$1->node_list_count]->bookname = $4;
  $1->node_list_count += 1;
  $1->node_list[$1->node_list_count] = (icm_node_list_entry_t *) 0;
  $$ = $1;
};

/*
 * section_part consists a repeated sequence of:
 *
 * [Begin ICM Section] <section name>                <-- section name must be previously referenced
 * [Derivation Method] <method type>                 <-- Required, <method type> is Lumped or Distributed, followed by:
 * rlgc_matrix                                       <-- defined below, require either rlgc or s-parameter, but not both
 * [ICM S-parameter]                                 <-- require either rlgc or s-parameter, but not both
 * File_name <file name>                             <-- Required, follows immediately
 * Port_assignment                                   <-- Required, follows immediately
 * <pin name> <input port name> <output port name>   <-- multiple lines of this until the next keyword
 * [End ICM Section] <section name>                  <-- section name must match
 */

section_end:
END_ICM_SECTION text_string newline {
  $$ = $2;
};

section_part: 
section_body section_end {
  if ($1->name != $2) _icm_msg(WARNING, &@1, "[Begin ICM Section] name '%s' does not match [End ICM Section] name '%s'.", $1->name, $2);
  $1->loc = @$;
  $$ = $1;
}|
section_body error section_end {
  _icm_msg(INFO, &@3, "Resuming parse after [End ICM Section] %s.",$3);
  yyerrok;
  $1->loc = @$;
  $$ = $1;
}|
section_body {
  YYLTYPE end_loc;
  end_loc.first_line = end_loc.last_line = @1.last_line + 1;
  end_loc.first_column = end_loc.last_column = 1;
  _icm_msg(ERROR, &end_loc, "Missing required [End ICM Section].  [End ICM Section] assumed.");
  $1->loc = @$;
  $$ = $1;
};

section_body:
BEGIN_ICM_SECTION text_string newline {
  $$ = _icm_new_icm_section();
  if (!strlen($2)) _icm_msg(ERROR, &@2, "A Section name is required.");
  $$->name = $2;
}|
section_body error newline {
  _icm_msg(INFO, &@3, "Resuming parse here.");
  /* yyerrok; */
  $$ = $1;
}|
section_body derivation_method newline { 
  if ($1->r || $1->l || $1->g || $1->c || $1->s_file) _icm_msg(ERROR, &@2, "[Derivation Method] is required before all section data.");
  if ($1->derivation_method) {
    _icm_msg(ERROR, &@2, "Redundant [Derivation Method] Keyword.");
  } else {
    if (!$2) {
      _icm_msg(WARNING, &@2, "Assuming '[Derivation Method] Lumped'.");
      $2 = _icm_string("Lumped");
    }
    $1->derivation_method = $2;
  }
  $$ = $1;
}|
section_body rlgc_matrix { 
  int i;
  if (!$1->derivation_method) _icm_msg(ERROR, &@2, "[Derivation Method] is required before all section data.");
  if ($1->s_file || $1->s_parameter_seen) {
    _icm_msg(ERROR, &@2, "The presence of [S-parameter] data precludes the use of the RLGC data.");
    _icm_free_icm_matrix($2);
  } else {
    
    /*
     * Scan the data for dimensionality.  Only the rows are reliable for determining this.
     */
    
    for (i = 0; i < $2->value_count; i++) {
      if ($2->rows < $2->values[i]->row) $2->rows = $2->values[i]->row;
      if ($2->cols < $2->values[i]->col) $2->cols = $2->values[i]->col;
    }
    
    switch ($2->representation[0]) {
    case 'F': 
      /* 
       * CHECK: Full matricies must have all entries above the upper diagonal
       */
      i = 0;
      while (i < $2->value_count) {
	int row = $2->values[i]->row;
	while (i < $2->value_count) {
	  if (row != $2->values[i]->row) break;
	  i += 1;
	}
	if ($2->values[i-1]->col != $2->cols) _icm_msg(ERROR, &@2, "%s row %d has %d entries, but %d are expected.", 
						       $2->type, row, $2->values[i-1]->col - row + 1, $2->cols - row + 1);
      }
      /* Yes, fall through */
    case 'D':
      /* Diagonal and Full matricies should be unambiguously square */
      if ($2->rows != $2->cols) _icm_msg(ERROR, &@2, "%s has bad dimensionality, it is %d X %d.", $2->type, $2->rows, $2->cols);
      break;
      /* Sparse matricies may not exhibit all columns, but there should never be more than rows */
    case 'S':
      if ($2->rows < $2->cols) _icm_msg(ERROR, &@2, "%s has bad dimensionality, it is %d X %d.", $2->type, $2->rows, $2->cols);
      if ($2->rows > $2->cols) _icm_msg(WARNING, &@2, "%s has questionable dimensionality, it is %d X %d.", $2->type, $2->rows, $2->cols);
      break;
    case 'B':
      /* Banded matricies bandwidth must be tested here, as only when fully read in can we do so */
      i = 0;
      while (i < $2->value_count) {
	int row = $2->values[i]->row;
	int mincol = row;
	int maxcol = row + $2->bandwidth;
	int entries;
	int count = 0;
	if (maxcol > $2->rows) maxcol = $2->rows;
	entries = maxcol - mincol + 1;
	while (i < $2->value_count) {
	  if (row != $2->values[i]->row) break;
	  count += 1;
	  i += 1;
	}
	if (count != entries) _icm_msg(ERROR, &@2, "%s has bad bandwidth at row %d, %d entr%s expected.", 
				       $2->type, row, entries, (1 < entries) ? "ies are" : "y is");
      }
      break;
    default: _icm_msg(ERROR, &@2, "FIXME: Internal fault, impossible representation '%s' detected.", $2->type);
    }

    /*
     * CHECK: ensure that all matricies have consistent row/colums within the section
     */
    if ($1->rows <= 0) $1->rows = $2->rows;
    if ($1->cols <= 0) $1->cols = $2->cols;
    if (($1->rows != $2->rows) || ($1->cols != $2->cols)) {
      _icm_msg(ERROR, &@2, "%s (%d by %d) is dimensionally inconsistent with Section '%s' (%d by %d).",
	       $2->type, $2->rows, $2->cols, $1->name, $1->rows, $1->cols);
    }

    switch (RLGC_TYPE($2)) {
    case RESISTANCE_MATRIX:  
      if ($1->r) {
        _icm_msg(ERROR, &@2, "Redundant %s", $2->type); 
        _icm_free_icm_matrix($2);
      } else { 
        $1->r = $2;
      } 
      break;
    case INDUCTANCE_MATRIX:  
      if ($1->l) {
        _icm_msg(ERROR, &@2, "Redundant %s", $2->type);
        _icm_free_icm_matrix($2);
      } else { 
        $1->l = $2;
      }
      break;
    case CONDUCTANCE_MATRIX:
      if ($1->g) {
         _icm_msg(ERROR, &@2, "Redundant %s", $2->type);
        _icm_free_icm_matrix($2);
      } else {
        $1->g = $2;
      }
      break;
    case CAPACITANCE_MATRIX:
      if ($1->c) {
        _icm_msg(ERROR, &@2, "Redundant %s", $2->type);
        _icm_free_icm_matrix($2);
      } else {
        $1->c = $2;
      }
      break;
    default:  _icm_msg(ERROR, &@2, "FIXME: Internal fault, impossible type '%s' detected.", $2->type);
    }
  }
  $$ = $1;
}|
section_body ICM_S_PARAMETER newline { 
  if ($1->r || $1->l || $1->g || $1->c) _icm_msg(ERROR, &@2, "The presence of RLGC data precludes the use of the %s data.", $2);
  if ($1->s_parameter_seen) {
    _icm_msg(ERROR, &@2, "Redundant %s Keyword.", $2);
  } else {
    $1->s_parameter_seen = 1;
  }
  $$ = $1;
}|
section_body File_name filename newline { 
  if ($1->r || $1->l || $1->g || $1->c) _icm_msg(ERROR, &@2, "The presence of RLGC data precludes the use of the %s subparameter.", $2);
  if ($1->s_file) {
    _icm_msg(ERROR, &@2, "Redundant %s subparameter.", $2);
  } else {
    $1->s_file = $3;
    switch (_icm_file_exists($3)) {
    case -1:
      _icm_msg(ERROR, &@3, "File '%s' does not exist in the current working directory.", $3);
      break;
    case 0:
      break;
    case 1:
      _icm_msg(ERROR, &@3, "File '%s' does not exist within directory '%s'.", $3, _icm_dir_name());
      break;
    default:
      _icm_msg(ERROR, &@3, "FIXME: Internal fault, impossible file existance check result.");
    }
  }
  $$ = $1;
}|
section_body Port_assignment newline port_assignment_list {
  if ($1->port_assignment) {
    int i;
    _icm_msg(ERROR, &@2, "Redundant '%s' subparameter.", $2); 
    for (i = 0; $4[i]; i++) _icm_free_icm_port_assignment_entry($4[i]);
    free((void *)$4);
  } else {
    for ($1->port_assignment_count = 0; $4[$1->port_assignment_count]; $1->port_assignment_count++);
    $1->port_assignment = $4;
  }
  $$ = $1;
};

derivation_method: 
DERIVATION_METHOD word { 
  $$ = choice(&@2, $2, "Lumped Distributed");
  if (!$$) {
    _icm_msg(ERROR, &@2, "'%s' is not one of: 'Lumped', or 'Distributed'.", $2);
  }
};

port_assignment_list: 
/* empty */ {
  $$ = (icm_port_assignment_entry_t **) calloc(1, (0 + 1) * sizeof(icm_port_assignment_entry_t *));
  $$[0] = (icm_port_assignment_entry_t *) 0;
}| port_assignment_list error newline {
  _icm_msg(WARNING, &@2, "Malformed port assignment entry.");
  _icm_msg(INFO, &@3, "Resuming parse here.");
}| port_assignment_list port_assignment_entry { 
  if (strlen($2->node) > 0) {
    int count;
    for (count = 0; $1[count]; count ++);
    $1[count] = $2;
    count +=1;
    $1 = (icm_port_assignment_entry_t **) realloc($1, (count + 1) * sizeof(icm_port_assignment_entry_t *));
    $1[count] = (icm_port_assignment_entry_t *) 0;
    $$ = $1;
  } else {
    _icm_msg(WARNING, &@2, "Malformed port assignment entry.");
    free($2);
  }
};

port_assignment_entry: 
word node_name newline {
  $$ = _icm_new_icm_port_assignment_entry();
  _icm_msg(DEBUG, &@$, "Capturing Port_assignment entry ('%s', '%s')", $1, $2);
  $$->port = $1;
  $$->node = $2;
};


/*
 * rlgc_matrix
 *
 * matrix_type                                <-- At most ONE matrix keyword (below) of each type per section, order irrelevant. 
 * [Bandwidth] <bandwidth>                    <-- ONLY if the <matrix format> is Banded_matrix
 * <matrix values>                            <-- defined below
 *
 * matrix_type:
 *
 * [Resistance Matrix] <matrix format>        <-- <matrix format> is Diagonal_matrix, Banded_matrix Sparse_matrix or Full_matrix
 * [Inductance Matrix] <matrix format>          
 * [Conductance Matrix] <matrix format>
 * [Capacitance Matrix] <matrix format>
 * 
 *
 * <matrix values>                            <-- depends on the <matrix format>
 *
 * <matrix format> == Diagonal_matrix         <-- diagonal elements only, one per line, both row and column are implied.
 * <matrix format> == Banded_matrix           <-- data consists of [row]s of <bandwith> columns for the upper diagonal elements
 * <matrix format> == Full_matrix             <-- data consists of [row]s of ALL columns for the upper diagonal
 * <matrix format> == Sparse_matrix           <-- data consists of [row]s of <column, value> tuples values for the upper diagonal
 *
 * error recovery tactics:
 *  none so far
 */

rlgc_matrix: 
RESISTANCE_MATRIX  matrix {
  $2->type = $1;
  $$ = $2;
}|
INDUCTANCE_MATRIX  matrix {
  $2->type = $1;
  $$ = $2;
}|
CONDUCTANCE_MATRIX matrix {
  $2->type = $1;
  $$ = $2;
}|
CAPACITANCE_MATRIX matrix {
  $2->type = $1;
  $$ = $2;
};

matrix: 
Sparse_matrix newline sparse_value {
  $$ = _icm_new_icm_matrix();
  $$->frequency=0.0;
  $$->representation = $1; 
  $$->values = $3;
  while($$->values[$$->value_count]) $$->value_count += 1;
  $$->loc = @$;
}|
Diagonal_matrix newline real_number_hint diagonal_value {
  $$ = _icm_new_icm_matrix();
  $$->frequency=0.0;
  $$->representation = $1; 
  $$->values = $4;
  while($$->values[$$->value_count]) $$->value_count += 1;
  $$->loc = @$;
}|
Banded_matrix newline BANDWIDTH integer newline normal_value {
  $$ = _icm_new_icm_matrix();
  $$->frequency=0.0;
  $$->representation = $1;
  if ($4 < 0) {
    _icm_msg(ERROR, &@4, "[Bandwidth] must be non-negative.");
    _icm_msg(WARNING, &@4, "Assuming '[Bandwidth] 0'.");
    $4 = 0;
  }
  $$->bandwidth = $4;
  $$->values = $6;
  while($$->values[$$->value_count]) $$->value_count += 1;
  $$->loc = @$;
}|
Full_matrix newline normal_value {
  $$ = _icm_new_icm_matrix();
  $$->frequency=0.0;
  $$->representation = $1; 
  $$->values = $3;
  while($$->values[$$->value_count]) $$->value_count += 1;
  $$->loc = @$;
};

/*
 * Sparse format:
 *
 * [Row] integer (starts at 1)
 * <col> <value>
 * <col> <value>
 * ...
 * [Row] integer+1
 * ...
 */

sparse_value: 
sparse_entry {
  icm_matrix_value_t **value = $1;
  if (value[0]) {
    if (value[0]->row != 1) _icm_msg(ERROR, &@1, "Expected '[Row] 1'.");
  } else {
    _icm_msg(ERROR, &@1, "[Row] %d has no values.", value[1]->row);    /* Not an error! value[1] is present when value[0] is nil */
  }
  $$ = $1;
}| sparse_value sparse_entry {
  int count;
  icm_matrix_value_t **value = $2;
  if (*value) {
    int duplicate_row = 0;
    for (count = 0; $1[count]; count ++) {
      if ($1[count]->row == value[0]->row) duplicate_row = 1;
    }
    if (duplicate_row) {
      int i;
      _icm_msg(ERROR, &@2, "Duplicate '[Row] %d'.", value[0]->row);
      for (i = 0; $2[i]; i++) _icm_free_icm_matrix_value($2[i]);
    } else {
      if (count && (($1[count - 1]->row + 1 ) != value[0]->row)) _icm_msg(ERROR, &@2, "Expected '[Row] %d'.", $1[count - 1]->row + 1);
      while (*value) {
	$1[count] = *value++;
	count += 1;
	$1 = (icm_matrix_value_t **) realloc($1, (count + 1) * (sizeof(icm_matrix_value_t *)));
      }
      $1[count] = (icm_matrix_value_t *) 0;
    }
  } else {
    _icm_msg(ERROR, &@2, "[Row] %d has no values.", value[1]->row);    /* Not an error! value[1] is present when value[0] is nil */
    _icm_free_icm_matrix_value(value[1]);
  }
  free((void *)$2);
  $$ = $1;
};

sparse_entry: 
ROW integer newline integer_hint {
  $$ = (icm_matrix_value_t **) calloc(1, (1 + 1) * sizeof(icm_matrix_value_t *));
  $$[0] = (icm_matrix_value_t *) 0;
  $$[1] = _icm_new_icm_matrix_value();
  if ($2 < 1) {
    _icm_msg(ERROR, &@2, "[Row] number '%d' must be greater than '0'.", $2);
    _icm_msg(WARNING, &@2, "Assuming '[Row] 1'.");
    $2 = 1;
  }
  $$[1]->row = $2;                               /* save for later */
}|
sparse_entry integer real_number newline integer_hint {
  int count;
  for (count = 0; $1[count]; count ++);
  if (count) {
    $$[count] = _icm_new_icm_matrix_value();
    $1[count]->row = $1[count - 1]->row;
  } else {
    $1[0] = $1[1];                               /* recapture previously saved starting row */
  }
  if ($1[count]->row > $2) {
    _icm_msg(ERROR, &@2, "Column '%d' must be greater than Row '%d'.", $1[count]->row);
    _icm_msg(WARNING, &@2, "Assuming Column '%d'.", $1[count]->row);
    $2 = $1[count]->row;;
  }
  $1[count]->col = $2;
  $1[count]->value = $3;

  count += 1;
  $1 = (icm_matrix_value_t **) realloc($1, (count + 1) * sizeof(icm_matrix_value_t *));
  $1[count] = (icm_matrix_value_t *) 0;

  $$ = $1;
}|
sparse_entry error newline integer_hint {
  _icm_msg(INFO, &@3, "Resuming parse here.");
  /* yyerrok; */
  $$ = $1;
};

/*
 * Diagonal format:
 *
 * <value>
 * <value>
 * ...
 */

diagonal_value: 
real_number newline real_number_hint {
  $$ = (icm_matrix_value_t **) calloc(1, (1 + 1) * sizeof(icm_matrix_value_t *));
  $$[0] = _icm_new_icm_matrix_value();
  $$[0]->row = 1;
  $$[0]->col = $$[0]->row;
  $$[0]->value = $1;
  $$[1] = (icm_matrix_value_t *) 0; 
}| diagonal_value real_number newline real_number_hint {
  int count;
  for (count = 0; $1[count]; count ++);
  $$[count] = _icm_new_icm_matrix_value();
  if (count < 1) {
    $1[count]->row = 1;
    $1[count]->col = 1;
  } else {
    $1[count]->row = $1[count-1]->row + 1;
    $1[count]->col = $1[count]->row;
  }
  $$[count]->value = $2;
  count += 1;
  $1 = (icm_matrix_value_t **) realloc($1, (count + 1) * sizeof(icm_matrix_value_t *));
  $1[count] = (icm_matrix_value_t *) 0;
  $$ = $1;
}|
diagonal_value error newline real_number_hint {
  _icm_msg(INFO, &@3, "Resuming parse here.");
  /* yyerrok; */
  $$ = $1;
};

/*
 * Normal format (banded or full)
 *
 * [Row] <integer
 * <value> <value> <value> ...
 * <value> <value> <value> ...
 * <value> <value> <value>
 * [Row] <integer
 * ...
 */

normal_value: 
normal_entry {
  icm_matrix_value_t **value = $1;
  if (value[0]) {
    if (value[0]->row != 1) _icm_msg(ERROR, &@1, "Expected '[Row] 1'.");
  } else {
    _icm_msg(ERROR, &@1, "[Row] %d has no values.", value[1]->row);    /* Not an error! value[1] is allocated even when value[0] is nil */
  }
  $$ = $1;
}| normal_value normal_entry { 
  int count;
  icm_matrix_value_t **value = $2;
  if (*value) {
    int duplicate_row = 0;
    for (count = 0; $1[count]; count++) {
      if ($1[count]->row == value[0]->row) duplicate_row = 1;
    }
    if (duplicate_row) {
      int i;
      _icm_msg(ERROR, &@2, "Duplicate '[Row] %d'.", value[0]->row);
      for (i = 0; $2[i]; i++) _icm_free_icm_matrix_value($2[i]);
    } else {
      if (count && (($1[count - 1]->row + 1) != value[0]->row)) _icm_msg(ERROR, &@2, "Expected '[Row] %d'.", $1[count - 1]->row + 1);
      while (*value) {
	$1[count] = *value++;
	count += 1;
	$1 = (icm_matrix_value_t **) realloc($1, (count + 1) * (sizeof(icm_matrix_value_t *)));
      }
      $1[count] = (icm_matrix_value_t *) 0;
    }
  } else {
    _icm_msg(ERROR, &@2, "[Row] %d has no values.", value[1]->row);    /* Not an error! value[1] is allocated even when value[0] is nil */
    _icm_free_icm_matrix_value(value[1]);
  }
  free((void *)$2);
  $$ = $1;
};

normal_entry: 
ROW integer newline real_number_hint {
  $$ = (icm_matrix_value_t **) calloc(1, (1 + 1) * sizeof(icm_matrix_value_t *));
  $$[0] = (icm_matrix_value_t *) 0; 
  $$[1] = _icm_new_icm_matrix_value();
  if ($2 < 1) {
    _icm_msg(ERROR, &@2, "[Row] number '%d' must be greater than '0'.", $2);
    _icm_msg(WARNING, &@2, "Assuming '[Row] 1'.");
    $2 = 1;
  }
  $$[1]->row = $2;                               /* save for later */
  $$[1]->col = $2;
}|
normal_entry real_number real_number_hint {
  int count;
  for (count = 0; $1[count]; count ++);
  if (count) {
    $1[count] = _icm_new_icm_matrix_value();
    $1[count]->row = $1[count - 1]->row;
    $1[count]->col = $1[count - 1]->col + 1;
  } else {
    $1[0] = $1[1];                               /* recapture previously saved starting row/column */
  }
  $1[count]->value = $2;

  count += 1;
  $1 = (icm_matrix_value_t **) realloc($1, (count + 1) * (sizeof(icm_matrix_value_t *)));
  $1[count] = (icm_matrix_value_t *) 0;

  $$ = $1;
}|
normal_entry newline real_number_hint {
  /* span multiple lines too */
}|
normal_entry error newline real_number_hint {
  _icm_msg(INFO, &@3, "Resuming parse here.");
  /* yyerrok; */
  $$ = $1;
};

%%

/*
 * Parse validation routines.
 * These are more complicated checks than should be accomodated in parser action routines.
 * They are typically invoked when a significant structure is successfully reduced.
 */

/*
 * Validate the ICM header struct upon it's completion
 */

static void validate_icm_header (icm_header_t *header)
{
  /*
   * ICM Header self-consistency tests:
   *
   * [ICM Ver] keyword is required.
   * [File Name] keyword is required.
   * [File Rev] keyword is required.
   * [Redistribution] keyword is required.
   * If redistribution type is 'Specific', then [Redistribution Text] is required.
   */
  
  if (!header->version)   {_icm_msg(ERROR, &header->loc, "Required keyword [ICM Ver] is missing.");}
  if (!header->file_name) {_icm_msg(ERROR, &header->loc, "Required keyword [File Name] is missing.");}
  if (!header->file_rev)  {_icm_msg(ERROR, &header->loc, "Required keyword [File Rev] is missing.");}
  if (!header->redist)    {_icm_msg(ERROR, &header->loc, "Required keyword [Redistribution] is missing.");}
  if ((header->redist == _icm_string("Specific")) && !header->redist_text)
    { 
      _icm_msg(ERROR, &header->loc, "'[Redistribution] Specific' requires the use of keyword [Redistribution Text].");
    }
}


/*
 * Validate an ICM Model Struct upon it's completion
 */

static void validate_icm_model (icm_model_t *model)
{
  
  /*
   * CHECK: SGR is required for SLM_general model types, must be omitted otherwise.
   * CHECK: Ref_impedance is only valid with SLM_* models
   * CHECK: A model must have either a Tree path, or a Nodal path.
   * CHECK: If swath parameters are used, all edges and swath pins must be defined as well.
   * CHECK: If Swath Parameters are present, at least TWO *_edges must be non-zero.
   */
  
  if (_icm_string("SLM_general") == model->type) {
    if (!model->sgr) {
      _icm_msg(ERROR, &model->loc, "ICM_model_type '%s' requires the presence of an SGR.", model->type);
    }
  } else {
    if (model->sgr) {
      _icm_msg(WARNING, &model->loc, "SGR should not be used with ICM_model_type '%s'.", model->type);
    }
  }

  if (!IS_SLM(model)) {
    if (model->ref_impedance) {
      _icm_msg(WARNING, &model->loc, "The Ref_impedance subparameter is valid only in SLM_* models");
    } 
  } else {
    if (!model->ref_impedance) {
      model->ref_impedance = 50.0;   /* use default value */
      _icm_msg(INFO, &model->loc, "Model '%s' defaulting to 'Ref_impedance = " G_FMT "' ohms.", model->name, model->ref_impedance);
    }
  }

  if ((model->tree_path == 0) && (model->node_path == 0)) { 
    _icm_msg(ERROR, &model->loc, "Either [Tree Path Description] or [Nodal Path Description] must be present.");
  } else {
    if (model->swath_present) {

      /*
       * CHECK: all [ICM Swath Parameters] *_edge subparameters are present.
       * CHECK: Top/Bottom_edge and/or Left/Right_edge are pairwise non-zero
       * CHECK: At least one pair is set.
       * CHECK: [ICM Swath Pin Numbers] is present.
       * CHECK: Swath *_edges do not exceed the physically available [ICM Swath Pin Numbers] row/col
       */    
      
      _icm_msg(DEBUG, &model->loc, "*_edge(s) are left %d, right %d, top %d, bottom %d", 
	       model->left_edge, model->right_edge, model->top_edge, model->bottom_edge);
      
      if (model->left_edge < 0) {
	_icm_msg(ERROR, &model->loc, "Model '%s' is missing [ICM Swath Parameters] subparameter Left_edge.", model->name);
	_icm_msg(WARNING, &model->loc, "Assuming 'Left_edge = 0'");
	model->left_edge = 0;
      }
      if (model->right_edge < 0) {
	_icm_msg(ERROR, &model->loc, "Model '%s' is missing [ICM Swath Parameters] subparameter Right_edge.", model->name);
	_icm_msg(WARNING, &model->loc, "Assuming 'Right_edge = 0'");
	model->right_edge = 0;
      }
      if (model->top_edge < 0) {
	_icm_msg(ERROR, &model->loc, "Model '%s' is missing [ICM Swath Parameters] subparameter Top_edge.", model->name);
	_icm_msg(WARNING, &model->loc, "Assuming 'Top_edge = 0'");
	model->top_edge = 0;
      }
      if (model->bottom_edge < 0) {
	_icm_msg(ERROR, &model->loc, "Model '%s' is missing [ICM Swath Parameters] subparameter Bottom_edge.", model->name);
	_icm_msg(WARNING, &model->loc, "Assuming 'Bottom_edge = 0'");
	model->bottom_edge = 0;
      }
      
      if (model->right_edge) {
	if (!model->left_edge) {
	  _icm_msg(ERROR, &model->loc, "Model '%s' [ICM Swath Parameters] Left_edge cannot be zero.", model->name);
	}
      } else {
	if (model->left_edge) {
	  _icm_msg(ERROR, &model->loc, "Model '%s' [ICM Swath Parameters] Right_edge cannot be zero.", model->name);
	}
      }
      
      if (model->top_edge) {
	if (!model->bottom_edge) {
	  _icm_msg(ERROR, &model->loc, "Model '%s' [ICM Swath Parameters] Bottom_edge cannot be zero.", model->name);
	}
      } else {
	if (model->bottom_edge) {
	  _icm_msg(ERROR, &model->loc, "Model '%s' [ICM Swath Parameters] Top_edge cannot be zero.", model->name);
	}
      }
      
      if (!model->top_edge && !model->bottom_edge && !model->left_edge && !model->right_edge) {
	_icm_msg(ERROR, &model->loc, "Model '%s' [ICM Swath Parameters] must have at least one pair of non-zero edges.", model->name);
      }
      
      if (!model->swath) {
	_icm_msg(ERROR, &model->loc, "Model '%s' is missing required keyword [ICM Swath Pin Numbers].", model->name);
      } else {
	
	if ((model->top_edge + model->bottom_edge) > model->swath_rows ) {
	  _icm_msg(ERROR, &model->loc, "Model '%s' top/bottom swath edges (%d) exceeds the available Swath Pin rows (%d).",
		   model->name, (model->top_edge + model->bottom_edge), model->swath_rows);
	} else if ((model->left_edge + model->right_edge) > model->swath_cols ) {
	  _icm_msg(ERROR, &model->loc, "Model '%s' left/right swath edges (%d) exceeds the available Swath Pin columns (%d).",
		   model->name, (model->left_edge + model->right_edge), model->swath_cols);
	}
      }
    }
  }
}


/*
 * Validate the ICM family struct upon it's completion
 *
 * NOTE: Recursion is used to validate the tree path.
 *       the 'walk' routine is called to examine each 'forked' path segment.
 */

static void walk_family_tree (icm_family_t *family, icm_model_t *model, 
			      icm_pin_map_t **pinmap, icm_tree_path_t **tree)
{
  
  /*
   * Recursively walk the Model's tree path looking for signs of trouble .....
   *
   * CHECK: All referenced Tree Path Pin Maps must be defined.
   * CHECK: Swathing must NOT be used if bounding the Tree path pinmap(s) are unordered.
   * CHECK: All Tree Path Pin Maps must have the same pin order
   * CHECK: All Tree Path Pin Maps must have the same number of pins.
   */
  
  while (*tree) {
    if (IS_FORK(*tree) && (*tree)->fork) {
      icm_tree_path_t *temp = (*tree)->fork;
      
      walk_family_tree(family, model, pinmap, &temp);
    } else {
      if (IS_PINMAP(*tree)) {
	icm_pin_map_t *prev_pinmap = *pinmap;
	
	*pinmap = icm_find_pin_map(family, (*tree)->name);
	if (!*pinmap) {
	  _icm_msg(ERROR, &(*tree)->loc, "Model '%s' references undefined Pin Map '%s'.", 
		   model->name, (*tree)->name);
	  *pinmap = prev_pinmap;
	} else {
	  if (model->swath_present && ((*pinmap)->pin_order == _icm_string("Unordered"))) {
	    _icm_msg(ERROR, &(*tree)->loc, "Model '%s' incorrectly uses swathing with Unordered Pin Map '%s'.", 
		     model->name, (*pinmap)->name);
	  }
	}
	if (prev_pinmap) {
	  if ((*pinmap)->pin_order != prev_pinmap->pin_order) {
	    _icm_msg(WARNING, &(*tree)->loc, "Model '%s' references Pin maps '%s' and '%s', which have an inconsistent pin order.", 
		     model->name, (*pinmap)->name, prev_pinmap->name);
	  } else {
	    if (prev_pinmap->pin_list_count != (*pinmap)->pin_list_count) {
	      _icm_msg(ERROR, &(*tree)->loc, "Model '%s' references Pin Maps '%s' and '%s', which have an inconsistent number of pins.", 
		       model->name, prev_pinmap->name, (*pinmap)->name);
	    }
	  }
	}
      }
    }
    (*tree) = (*tree)->next;
  }
}

static void validate_icm_family (icm_family_t *family)
{
  /*
   * ICM Family self-consistency tests:
   * 
   * General tests ---
   *
   * CHECK: [Manufacturer] keyword is required.
   * CHECK: [ICM Family Description] keyword is required.
   * CHECK: There must be a non-empty model list.
   */

  if (!family->manufacturer) _icm_msg(ERROR, &family->loc, "Missing [Manufacturer] keyword.");
  if (!family->description) _icm_msg(ERROR, &family->loc, "Missing [ICM Family Description] keyword.");
  if (!family->models) {
    _icm_msg(ERROR, &family->loc, "Missing [ICM Model List] keyword.");
  } else {
    
    /*
     * Model tests ---
     *
     * CHECK: All entries of the model list must be defined.
     * CHECK: Recursively walk the Model's tree path looking for signs of trouble .....
     * CHECK: All referenced Node Path Node Maps must be defined.
     * CHECK: nodes defined by Node Maps are referenced within the enclosing Node Path.
     *
     * TODO??? Not checkable? Treepath pinmaps must share the same pin ordering.
     */
    
    int i;
    
    for (i = 0; i < family->model_count; i++) {
      icm_model_t *model = family->models[i];
      icm_tree_path_t *tree_path = model->tree_path;
      icm_node_path_t **node_path = model->node_path;

      if (!model->type) {
	_icm_msg(ERROR, &model->loc, "Model '%s' was not defined by a [Begin ICM Model] keyword.", model->name);

      } else if (tree_path) {
	icm_pin_map_t *pinmap = (icm_pin_map_t *)0;
	
	walk_family_tree(family, model, &pinmap, &tree_path);
	
      } else if (node_path) {
	int node_path_count = model->node_path_count;
	icm_node_map_t *node_map;
	int j;

	for (j = 0; j < node_path_count; j++) {
	  if (IS_NODEMAP(node_path[j])) {
	    node_map = icm_find_node_map(family, node_path[j]->name);
	    if (!node_map) {
	      _icm_msg(ERROR, &node_path[j]->loc, "Model '%s' references undefined Node Map '%s'.", model->name, node_path[j]->name);
	    } else {
	      int k;
	      icm_node_list_entry_t **node_list = node_map->node_list;
	      for (k = 0; k < node_map->node_list_count; k++) {
		int l;
		int connected = 0;
		char *node_name = node_list[k]->nodename;
		for (l = 0; l < node_path_count; l++) {
		  if (IS_N_SECTION(node_path[l])) {
		    int m;
		    for (m = 0; m < node_path[l]->node_count; m++) {
		      if (node_name == node_path[l]->nodes[m]) {
			connected = 1;
			_icm_msg(DEBUG, &node_path[l]->loc, "Node Map '%s' node '%s' is connected to N_section '%s' in Model '%s'.",
				 node_path[j]->name, node_name, node_path[l]->name, model->name);
		      }
		    }
		  }
		}
		if (!connected) {
		  _icm_msg(WARNING, &node_path[j]->loc, "Node Map '%s' node '%s' is not connected to any N_section in Model '%s'.",
			   node_path[j]->name, node_name, model->name);
		}
	      }
	    }
	  }
	}
      }
    }
  }
}


/*
 * Validate the ICM Section Struct upon it's completion
 */

static void validate_icm_section (icm_section_t *section)
{
  /*
   * CHECK: Ensure a derivation method was established.  Default to Lumped.
   * CHECK: a Section may use either RLGC OR S-parameter data, not both, and must have at least one.
   * CHECK: if a section is 'distributed' it MUST have both inductance and capacitance matricies (and NO S-parameter data).
   * CHECK: if a section is 'lumped' it may have any matrix (at least one of them though)
   * CHECK: An S-Parameter section must have a File_name and a Port_assignment
   */
  
  /* Not sure I should even check these, more of a simulator issue */
  /* TODO: RLGC section resistance matrix is optional */
  /* TODO?: RLGC section resistance matrix, if present, should have all of its diagonal elements */
  /* TODO?: All SLM model conductors must have non-zero L and C values in at least one section */
  /* TODO?: MLM model RLGC section capactitance matricies off diagonal elements should be negative or zero */
  
  
  if (!(section->derivation_method)) {
    _icm_msg(ERROR, &section->loc, "Missing required [Derivation Method] Keyword.  Assuming '[Derivation Method] Lumped'.");
    section->derivation_method = _icm_string("Lumped");
  }
  
  if (section->s_parameter_seen) {
    if (!section->s_file) {
      _icm_msg(ERROR, &section->loc, "Section '%s' [S-parameter] File_name is missing.", section->name);
    }
    if (!section->port_assignment) {
      _icm_msg(ERROR, &section->loc, "Section '%s' [S-parameter] Port_assignment is missing.", section->name);
    }
  }
  
  if (section->s_parameter_seen && !section->s_file) {
    _icm_msg(ERROR, &section->loc, "Section '%s' has no S-parameter data.", section->name);
  }
  
  if (!(section->s_parameter_seen || section->r || section->l || section->g || section->c)) {
    _icm_msg(ERROR, &section->loc, "Section '%s' has no RLGC data.", section->name);
  }
  
  if (section->s_parameter_seen && section->s_file) {
    if ((section->r || section->l || section->g || section->c)) {
      _icm_msg(ERROR, &section->loc, "Section '%s' erroneously uses both RLGC and S-parameter data.", section->name);
    }
    if (IS_DISTRIBUTED(section)) {
      _icm_msg(ERROR, &section->loc, "Section '%s' S-parameter data must use the Lumped derivation method.", section->name);
    }
  }
  
  if (IS_RLGC(section)) {
    if (IS_LUMPED(section)) {
      if (!(section->r || section->l || section->g || section->c)) {
	_icm_msg(ERROR, &section->loc, "Section '%s' Lumped derivation requires at least one RLGC Matrix.", section->name);
      }
    } else { /* Distributed RLGC */
      
      if (!section->l) _icm_msg(ERROR, &section->loc, "Section '%s' must have an [Inductance Matrix] since it's derivation method is distributed.", 
				section->name);
      if (!section->c) _icm_msg(ERROR, &section->loc, "Section '%s' must have a [Capacitance Matrix] since it's derivation method is distributed.", 
				section->name);
    }
  }
}


/*
 * Validate the ICM struct upon it's completion
 *
 * NOTE: Recursion is used to validate the tree path.
 *       the 'walk' routine is called to examine each 'forked' Model path segment.
 */

static void walk_icm_tree (icm_t *icm, icm_model_t *model, 
			   icm_section_t **section, icm_tree_path_t **tree)
{

  /*
   * CHECK: all referenced sections must be defined.
   * CHECK: S-parameter data may not be used with [Tree Path Description]
   * CHECK: all RLGC sections within a tree must have exactly the same dimensionality.
   * CHECK: SLM_* models must have purely diagonal RLGC matrix sections
   * CHECK: 'Mult' path sections must have 'Lumped' derivations
   * CHECK: 'Len' path sections must have 'Distributed' derivations
   * CHECK: Treepath pinmaps must have the same pin count as the Section RLGC diagonals
   *        They may have more pins if the model is swathed and the pinmap(s) are ordered
   *        They may not have pins than the matrix diagonal.
   */

  while (*tree) {
    if (IS_FORK(*tree) && (*tree)->fork) {
      icm_tree_path_t *temp = (*tree)->fork;
      walk_icm_tree(icm, model, section, &temp);
    } else {
      if (IS_SECTION(*tree)) {
	icm_section_t *prev_section = *section;
	*section = icm_find_section(icm, (*tree)->name);
	if (!*section) {
	  _icm_msg(ERROR, &(*tree)->loc, "Model '%s' references non-existant Section '%s'.", model->name, (*tree)->name);
	  *section = prev_section;
	} else {
	  if (IS_S_PARAMETER(*section)) {
	    _icm_msg(ERROR, &(*tree)->loc, "Model '%s' incorrectly references S-parameter Section '%s' using [Tree Path Description].",
		     model->name, (*section)->name);
	  } else {
	    if (prev_section) {
	      if ((prev_section->rows != (*section)->rows) || (prev_section->cols != (*section)->cols)) {
		_icm_msg(ERROR, &(*tree)->loc, "Model '%s' section '%s' is dimensionally inconsistent with section '%s'.",
			 model->name, (*section)->name, prev_section->name);
	      }
	    }
	    if (IS_SLM(model)) {
	      if (!section_is_diagonal(*section)) {
		_icm_msg(ERROR, &(*tree)->loc, "Model '%s' is of type '%s', but references non-diagonal RLGC Section '%s'.", 
			 model->name, model->type, (*section)->name);
	      }
	    }
	    if (IS_MULT(*tree) && IS_DISTRIBUTED(*section)) {
	      _icm_msg(ERROR, &(*tree)->loc, "Model '%s' incorrectly uses 'Mult=' with Distributed RLGC section '%s'.", 
		       model->name, (*section)->name);
	    }
	    if (IS_LEN(*tree) && IS_LUMPED(*section)) {
	      _icm_msg(ERROR, &(*tree)->loc, "Model '%s' incorrectly uses 'Len=' with Lumped RLGC section '%s'.", 
		       model->name, (*section)->name);
	    }
	  }
	}
      }
      
      /*
       * If we've found a pinmap, check it against a path Section.
       * ANY tree path section will do, they must all have the same dimensionality
       * Find one in the main path if we haven't already got one.
       *
       * This actually does something reasonable, we will typically already have
       * the nearest adjacent Section in hand, and if we have none, then the
       * very first section in the Tree path is the most appropriate one to examine.
       *
       * NOTE: pinmaps have already been checked for existance.
       *       no need to complain again if we can't find it.
       */
      
      if (IS_PINMAP(*tree)) {
	icm_pin_map_t *pinmap = icm_find_pin_map(icm->family, (*tree)->name);
	if (!*section) {
	  icm_tree_path_t *temp = model->tree_path;
	  while(temp && !IS_SECTION(temp)) temp = temp->next;
	  if (!temp) {
	    _icm_msg(ERROR, &(*tree)->loc, "Model '%s' pinmap '%s' does not connect to any Section.", 
		     model->name, pinmap->name);
	  } else {
	    *section = icm_find_section(icm, temp->name);
	  }
	}

	/*
	 * determine if the pinmap has the right number of pins for the RLGC section
	 * and if not, whether or not swathing can and is being used.
	 */

	if (*section && pinmap) {
	  if (IS_RLGC(*section)) {
	    if ((pinmap->pin_list_count > (*section)->rows)) {
	      if (model->swath_present) {
		if (IS_UNORDERED(pinmap)) {
		  _icm_msg(ERROR, &(*tree)->loc, "Model '%s' is swathed, but incorrectly uses Unordered pinmap '%s'.",
			   model->name, pinmap->name);
		}
	      } else {
		_icm_msg(ERROR, &(*tree)->loc, "Section '%s' has fewer rows than pinmap '%s', but Model '%s' is not swathed.",
			 (*section)->name, pinmap->name, model->name);
	      }
	    } else if ((pinmap->pin_list_count < (*section)->rows)) {
	      _icm_msg(ERROR, &(*tree)->loc, "Section '%s' has more rows than pinmap '%s' has pins.",
		       (*section)->name, pinmap->name);
	    } else {
	      /* ahh, exactly the right number of rows/pins */
	    }
	  }
	}
      }
    }
    *tree = (*tree)->next;
  }
}


static void validate_icm (icm_t *icm)
{
  /* 
   * CHECK: walk the model's tree path, looking for trouble...see previous routine.
   * CHECK: all referenced N_sections must be defined.
   * CHECK: S-parameter models must use nodal sections referencing s-parameter data
   * CHECK: All Nodal paths must use exclusively RLGC or s-parameter, not both
   * CHECK: SLM_* models may not use S-parameter sections
   * CHECK: 'Mult' RLGC N_sections must have 'Lumped' derivations
   * CHECK: 'Len' RLGC N_sections must have 'Distributed' derivations
   * CHECK: 'Len' may not be used with S-parameter N_sections (which are always Lumped)
   * CHECK: N_sections using RLGC matricies must have even numbers of nodes, the RLGC diagonals are half this value
   * CHECK: All defined Sections, pinmaps and nodelists are actually referenced.
   * CHECK: N_sections referencing S-parameter data must match all nodes of the Section's Port_assignment.
   */

  /* TODO: left half names of an n_section are definitions, right half references.
     portmap names may match either half (all must be found), and are thus either definitions or references.
     All references must be defined either by a portmap or an n_section. 
     Not all definitions need be referenced. */
 
  if (!icm->family) {
    _icm_msg(ERROR, &icm->loc, "Missing ICM Family (due to previous errors?)");
  } else {
    icm_model_t **models = icm->family->models;
    int model_count = icm->family->model_count;
    
    int i, j;
    
    for (i = 0; i < model_count; i++) {
      icm_model_t * model = models[i];
      icm_tree_path_t *tree = model->tree_path;
      icm_node_path_t **node_path = model->node_path;
      icm_section_t *section = (icm_section_t *)0;
      icm_section_t *prev_section = (icm_section_t *)0;

      if (tree) {
	walk_icm_tree(icm, model, &section, &tree);
      } else if (node_path) {
	_icm_msg(DEBUG, &node_path[0]->loc, "Examining Model '%s' Nodal Paths.  There are a total of %d",
		 model->name, model->node_path_count);
	for (j = 0; j < model->node_path_count; j++) {
	  if (IS_N_SECTION(node_path[j])) {
	    prev_section = section;
	    section = icm_find_section(icm, node_path[j]->name);
	    if (!section) {
	      section = prev_section;
	      _icm_msg(ERROR, &node_path[j]->loc, "Model '%s' references non-existant N_section '%s'.", model->name, node_path[j]->name);
	    } else {
	      if (_icm_string("S-parameter") == model->type && IS_RLGC(section)) {
		_icm_msg(ERROR, &node_path[j]->loc, "Model '%s' is of type '%s', but incorrectly references RLGC N_section '%s'.", 
			 model->name, model->type, section->name);
	      }
	      if (prev_section) {
		if (IS_RLGC(prev_section) != IS_RLGC(section)) {
		  _icm_msg(ERROR, &node_path[j]->loc, "Model '%s' sections '%s' and '%s' mix RLGC and S-parameter data.",
			   model->name, prev_section->name, section->name);
		}
	      }
	      if (IS_RLGC(section)) {
		_icm_msg(DEBUG, &node_path[j]->loc, "Examining RLGC '%s', referencing N_section[%d] has %d nodes.",
			 section->name, j, node_path[j]->node_count);
		if (IS_MULT(node_path[j]) && IS_DISTRIBUTED(section)) {
		  _icm_msg(ERROR, &node_path[j]->loc, "Model '%s' incorrectly uses 'Mult=' with Distributed RLGC section '%s'.", 
			   model->name, section->name);
		}
		if (IS_LEN(node_path[j]) && IS_LUMPED(section)) {
		  _icm_msg(ERROR, &node_path[j]->loc, "Model '%s' incorrectly uses 'Len=' with Lumped RLGC section '%s'.", 
			   model->name, section->name);
		}
		if (node_path[j]->node_count % 2) {
		  _icm_msg(ERROR, &node_path[j]->loc, "Model '%s' N_section '%s' must have an even number of nodes since it uses an RLGC Section",
			   model->name, prev_section->name, section->name);
		} else {
		  if (section->rows != (node_path[j]->node_count / 2)) {
		    int need =  section->rows * 2;
		    _icm_msg(ERROR, &node_path[j]->loc, "Model '%s' N_section '%s' requires %d node%s because the RLGC Section has %d rows.",
			     model->name, section->name, need, (1 != need) ? "s" : "", section->rows);
		  }
		}
	      } else {  /* Necessarily an S-parameter section */
		int k, l;

		_icm_msg(DEBUG, &node_path[j]->loc, "Examining S-parameter section '%s', referencing N_section[%d] has %d nodes.",
			 section->name, j, node_path[j]->node_count);
		
		if (section->port_assignment_count != node_path[j]->node_count) {
		  _icm_msg(ERROR, &node_path[j]->loc, "N_section '%s' node count (%d) does not match Port_assignment Node count (%d).",
			   section->name, node_path[j]->node_count, section->port_assignment_count);
		}

		/*
		 * See if the N_section nodes are all present in the Port_assignment
		 */

		for (k = 0;  k < node_path[j]->node_count; k++) {
		  int found = 0;
		  for (l = 0; l < section->port_assignment_count; l++) {
		    if (node_path[j]->nodes[k] == section->port_assignment[l]->node) {
		      found = 1;
		    }
		  }
		  if (!found) {
		    _icm_msg(ERROR, &node_path[j]->loc, "N_section '%s' Node '%s' is not present in the [S-parameter] Port_assignment.",
			     section->name, node_path[j]->nodes[k]);
		  }
		}

		/*
		 * Do the reverse test too. They must all match.
		 */

		for (l = 0; l < section->port_assignment_count; l++) {
		  int found = 0;
		  for (k = 0;  k < node_path[j]->node_count; k++) {
		    if (node_path[j]->nodes[k] == section->port_assignment[l]->node) {
		      found = 1;
		    }
		  }
		  if (!found) {
		    _icm_msg(ERROR, &node_path[j]->loc, "Section '%s' Port_assignment Node '%s' is not present in the N_section.",
			     section->name, section->port_assignment[l]->node);
		  }
		}


		if (IS_SLM(model)) {
		  _icm_msg(ERROR, &node_path[j]->loc, "Model '%s' is of type '%s', but incorrectly references S-parameter N_section '%s'.", 
			   model->name, model->type, section->name);
		}
		if (IS_LEN(node_path[j])) {
		  _icm_msg(ERROR, &node_path[j]->loc, "Model '%s' must not use 'Len=' with S-parameter N_section '%s'.", 
			   model->name, section->name);
		}
	      }
	    }
	  }
	}
      }
    }
    /* 
     * CHECK: Ensure that all Family pinmaps and nodemaps have been referenced at least once.
     */
    {
      int i;
      icm_pin_map_t **pinmap = icm->family->pin_maps;
      icm_node_map_t **nodemap = icm->family->node_maps;

      for (i = 0; i < icm->family->pin_map_count; i++) {
	if (!pinmap[i]->referenced) _icm_msg(WARNING, &pinmap[i]->loc, "Pin Map '%s' was defined but not used.", pinmap[i]->name);
      }

      for (i = 0; i < icm->family->node_map_count; i++) {
	if (!nodemap[i]->referenced) _icm_msg(WARNING, &nodemap[i]->loc, "Node Map '%s' was defined but not used.", nodemap[i]->name);
      }
    }
  }
  /*
   * CHECK: Ensure that all ICM Sections have been referenced at least once.
   */
  {
    int i;
    icm_section_t **section = icm->sections;

    for (i = 0; i < icm->section_count; i++) {
      if (!section[i]->referenced) _icm_msg(WARNING, &section[i]->loc, "Section '%s' was defined but not used.", section[i]->name);
    }
  }
}


/*
 * utility functions
 */

static char *choice (YYLTYPE *loc, char *in, char *choices)
{
  char *c = choices;
  char *p = choices;
  char *ans = in;

  if (!ans) return((char *)0);

  while (*p) {
    while (*c && (*c != ' ')) { c++; }
    if (!dyn_strncasecmp(p, ans, c-p)) {
      if (!strncmp(p, ans, c-p)) {
	return(ans);
      } else {
	char *temp = dyn_strndup(p, c-p+1);
	temp[c-p] = '\0';
	_icm_msg(ERROR, loc, "'%s' is not correctly cased.  Use '%s'.", ans, temp);
	ans = _icm_string(ans);
	free(temp);
	return(ans);
      }
    }
    while (*c && (*c == ' ')) { c++; }
    p = c;
  }
  return(0);
}


/*
 * determine if a section, and if a matrix is diagonal or not
 *
 * S_parameter sections return -1 (an error)
 * otherwise, 0 is returned if non-diagonal, 1 if diagonal
 *
 * Empty matricies are considered to be diagonal.
 *
 */

static int section_is_diagonal (icm_section_t *section)
{
  int is_diagonal = 1;

  if (section->s_file) return -1;

  if (section->r && !matrix_is_diagonal(section->r)) {
    _icm_msg(INFO, &section->r->loc, "Section '%s' '%s' is non-diagonal.", section->name, section->r->type);
    is_diagonal = 0;
  } 
  if (section->l && !matrix_is_diagonal(section->l)) {
    _icm_msg(INFO, &section->l->loc, "Section '%s' '%s' is non-diagonal.", section->name, section->l->type);
    is_diagonal = 0;

  }
  if (section->g && !matrix_is_diagonal(section->g)) {
    _icm_msg(INFO, &section->g->loc, "Section '%s' '%s' is non-diagonal.", section->name, section->g->type);
    is_diagonal = 0;

  }
  if (section->c && !matrix_is_diagonal(section->c)) {
    _icm_msg(INFO, &section->c->loc, "Section '%s' '%s' is non-diagonal.", section->name, section->c->type);
    is_diagonal = 0;

  }
  return is_diagonal;
  
}

static int matrix_is_diagonal (icm_matrix_t *matrix)
{
  int i;
  icm_matrix_value_t **value = matrix->values;

  /* 
   * look for any non-zero off-diagonal elements
   */

  for (i = 0; i < matrix->value_count; i++) {
    if (((matrix->values[i])->row != (matrix->values[i])->col) && (0.0 != (matrix->values[i])->value)) return 0;
  }

  return 1;
}
