#ifndef ICM_IMPL_H
#define ICM_IMPL_H

/*
 ******************************************************************************
 *
 * This file is a part of the IBIS ICM Parser.
 * Copyright (C) 2003-2005 Kelly C. Green <kgreen22@mindspring.com>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 ******************************************************************************
 *
 * $URL: file:///home/kgreen/ibis-svn-root/icmchk/trunk/src/icm-impl.h $
 *
 * $Id: icm-impl.h 109 2005-02-24 07:43:51Z kgreen $
 *
 ******************************************************************************
 */

/*
 * This is the private include file for the icm parser.
 * It's only for use by modules implementing the parser itself.
 */

#include "icm.h"
#include "dynstr.h"

/*
 * The Lexical analyzer's input file descriptor
 */

extern FILE *yyin;

/*
 * Lexical analyzer "hint"s
 */

enum hint_e { H_DEFAULT,
              H_STRING,
              H_BLOCK,
              H_FNAME,
              H_WORD,
	      H_NODE_NAME,
              H_NUMBER,
              H_SGR
};

typedef enum hint_e hint_t;

/*
 * prototypes for assorted parser internal functions
 */

extern int yyparse                PROTO((void));
extern int icm_1_0_parse          PROTO((void));

extern void yylex_init            PROTO((void));
extern void yylex_reinit_as       PROTO((char *version));
extern void yylex_finish          PROTO((void));
extern int  yylex                 PROTO((void));
extern void yylex_flush_keyword   PROTO((void));
extern void yylex_flush_line      PROTO((void));
extern void yylex_disable_subp    PROTO((void));
extern void yylex_hint            PROTO((hint_t type));
extern char yylex_x_char          PROTO((char cchar));
extern void yylex_enable_x_char   PROTO((void));

/*
 * Parser internal interface wrappers
 */

extern char *_icm_version          PROTO((void));
extern void _icm_emit_result       PROTO((icm_t *icm));
extern int _icm_file_exists        PROTO((char *name));
extern char *_icm_file_name        PROTO((void));
extern char *_icm_dir_name         PROTO((void));

/*
 * error handlers used by the parser
 */

#ifdef yyerror
#undef yyerror
#endif

#define yyerror(msg)  _icm_msg(ERROR, &yylloc, msg)
  
void _icm_msg                   PROTO((level_t level, YYLTYPE *loc, char *s, ...));

/*
 * private structure creation/deletion methods
 */

char                        * _icm_string                        PROTO((char *string));
icm_t                       * _icm_new_icm                       PROTO((void));
icm_header_t                * _icm_new_icm_header                PROTO((void));
icm_family_t                * _icm_new_icm_family                PROTO((void));
icm_model_t                 * _icm_new_icm_model                 PROTO((void));
icm_tree_path_t             * _icm_new_icm_tree_path             PROTO((void));
icm_node_path_t             * _icm_new_icm_node_path             PROTO((void));
icm_swath_entry_t           * _icm_new_icm_swath_entry           PROTO((void));
icm_pin_map_t               * _icm_new_icm_pin_map               PROTO((void));
icm_pin_list_entry_t        * _icm_new_icm_pin_list_entry        PROTO((void));
icm_node_map_t              * _icm_new_icm_node_map              PROTO((void));
icm_node_list_entry_t       * _icm_new_icm_node_list_entry       PROTO((void));
icm_section_t               * _icm_new_icm_section               PROTO((void));
icm_matrix_t                * _icm_new_icm_matrix                PROTO((void));
icm_matrix_value_t          * _icm_new_icm_matrix_value          PROTO((void));
icm_port_assignment_entry_t * _icm_new_icm_port_assignment_entry PROTO((void));

void _icm_free_icm_header                PROTO((icm_header_t *header));
void _icm_free_icm_family                PROTO((icm_family_t *family));
void _icm_free_icm_model                 PROTO((icm_model_t *model));
void _icm_free_icm_tree_path             PROTO((icm_tree_path_t *tree_path));
void _icm_free_icm_node_path             PROTO((icm_node_path_t *node_path));
void _icm_free_icm_swath_entry           PROTO((icm_swath_entry_t *swath_entry));
void _icm_free_icm_pin_map               PROTO((icm_pin_map_t *pin_map));
void _icm_free_icm_pin_list_entry        PROTO((icm_pin_list_entry_t *pin_list_entry));
void _icm_free_icm_node_map              PROTO((icm_node_map_t *node_mape));
void _icm_free_icm_node_list_entry       PROTO((icm_node_list_entry_t *node_list_entry));
void _icm_free_icm_section               PROTO((icm_section_t *section));
void _icm_free_icm_matrix                PROTO((icm_matrix_t *matrix));
void _icm_free_icm_matrix_value          PROTO((icm_matrix_value_t *matrix_value));
void _icm_free_icm_port_assignment_entry PROTO((icm_port_assignment_entry_t *port_assignment_entry));


#endif
