#ifndef INCLUDES_H
#define INCLUDES_H

/*
 ******************************************************************************
 *
 * This file is a part of the IBIS ICM Parser.
 * Copyright (C) 2003-2005 Kelly C. Green <kgreen22@mindspring.com>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 ******************************************************************************
 *
 * $URL: file:///home/kgreen/ibis-svn-root/icmchk/trunk/src/includes.h $
 *
 * $Id: includes.h 119 2005-03-10 05:18:20Z kgreen $
 *
 ******************************************************************************
 */

/*
 * Try to include the host configuration.
 * If the host is a windows system, we fake it.
 * Otherwise require the use of the configure tool to create it.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#else
# ifdef _WIN32
#
#  define PACKAGE_VERSION    "1.1.0"
#  define PACKAGE_NAME       "icmchk"
#
#  define STDC_HEADERS   1
#  define HAVE_STRING_H  1
#  define HAVE_CTYPE_H   1
#  define HAVE_MATH_H    1
#  define HAVE_ASSERT_H  1
#  define HAVE_FREE      1
#  define HAVE_MALLOC    1
#  define HAVE_REALLOC   1
#  define PROTOTYPES     1
#
/*
 * Disable warnings about deprecated functions.
 * Microsoft has deprecated a number of Standard C89/CLIB Functions.
 * This requires some shiny brass ones and a total disregard for cross
 * platform program portability.
 */
#
#  pragma warning(disable : 4996)
#
/*
 * disable the warning about an unreferenced label.
 * it's 'yyerrorlab' and is in yacc generated code.
 */
#
#  pragma warning(disable : 4102)
#
/*
 * Disable /Wp64 portability checks, which are on by default in
 * Visual C++ Express, and cause the compiler to warn about
 * possible loss of precision when converting 64 bit char
 * pointers to 32 bit ints.  
 *
 * In practice, these warnings exist only to flag possible 64 bit
 * portability issues.  All that are fixable have been fixed.  The
 * residue lies in generated C code by yacc and lex over which 
 * there is no direct control, and in any case either involves
 * pointer differences within buffers that are only 8k, or in
 * the fread() function return value, which is also given only 
 * an 8k buffer and thus cannot excede that value.
 *
 * You can re-enable these warning if you enjoy seeing them.
 */
#
#  pragma warning(disable : 4244)
#  pragma warning(disable : 4267)
#
# else
#  error "use -D HAVE_CONFIG_H and create 'config.h' with the 'configure' script."
# endif
#endif

/*
 * Include common includes
 */

#include <stdio.h>

#if STDC_HEADERS
# include <stdlib.h>
# include <stddef.h>
#else
# if HAVE_STDLIB_H
#  include <stdlib.h>
# endif
#endif

#if HAVE_STRING_H
# if !STDC_HEADERS && HAVE_MEMORY_H
#  include <memory.h>
# endif
# include <string.h>
#endif

#if HAVE_CTYPE_H
# include <ctype.h>
#endif

#if HAVE_MATH_H
# include <math.h>
#endif

#if HAVE_ASSERT_H
# include <assert.h>
#endif

#include <stdarg.h>

/*
 * Define a macro to conditionalize prototype declarations
 * Automake can convert function definitions to K&R format
 * if it needs to.
 */

#if __STDC__ && PROTOTYPES
# define PROTO(args)  args
#else
# define PROTO(args)  ()
#endif

#ifdef USE_DMALLOC
#include <dmalloc.h>
#endif

#endif
